/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    testMachineLearning.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Regression tests for machine learning classes.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <cmath>

#include <cxcore.h>

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

using namespace std;

// prototypes ---------------------------------------------------------------

void testLogistic();
void testConditionalGaussian();
void testSufficientStatistics();

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n"; 
    cerr << "USAGE: ./testMachineLearning [OPTIONS]\n";
    cerr << "OPTIONS:\n"
         << "  -logistic         :: test logistic regression\n"
         << "  -gaussian         :: test gaussians\n"
         << "  -suffstats        :: test sufficient statistics\n"
         << SVL_STANDARD_OPTIONS_USAGE
         << endl;
}

int main(int argc, char *argv[])
{
    if (argc <= 1) {
        usage();
        return -1;
    }

    char **args = argv + 1;
    while (--argc > 0) {
        SVL_PROCESS_STANDARD_OPTIONS(args, argc)
        else if (!strcmp(*args, "-logistic")) {
            testLogistic();
        } else if (!strcmp(*args, "-gaussian")) {
            testConditionalGaussian();
        } else if (!strcmp(*args, "-suffstats")) {
            testSufficientStatistics();
	} else {
	    cerr << "ERROR: unrecognized option " << *args << endl;
	    usage();
	    return -1;
	}
	args++;
    }

    svlCodeProfiler::print(cerr);
    return 0;
}

// tests ------------------------------------------------------------------

void testLogistic()
{
    SVL_ASSERT(false);
}

void testConditionalGaussian()
{
    CvMat *mu = cvCreateMat(3, 1, CV_32FC1);
    CvMat *sigma = cvCreateMat(3, 3, CV_32FC1);
    
    CV_MAT_ELEM(*mu, float, 0, 0) = 1.0;
    CV_MAT_ELEM(*mu, float, 1, 0) = 2.0;
    CV_MAT_ELEM(*mu, float, 2, 0) = 3.0;

    CV_MAT_ELEM(*sigma, float, 0, 0) = 1.0;
    CV_MAT_ELEM(*sigma, float, 0, 1) = 0.1;
    CV_MAT_ELEM(*sigma, float, 0, 2) = 0.2;
    CV_MAT_ELEM(*sigma, float, 1, 0) = CV_MAT_ELEM(*sigma, float, 0, 1);
    CV_MAT_ELEM(*sigma, float, 1, 1) = 1.0;
    CV_MAT_ELEM(*sigma, float, 1, 2) = -0.5;
    CV_MAT_ELEM(*sigma, float, 2, 0) = CV_MAT_ELEM(*sigma, float, 0, 2);
    CV_MAT_ELEM(*sigma, float, 2, 1) = CV_MAT_ELEM(*sigma, float, 1, 2);
    CV_MAT_ELEM(*sigma, float, 2, 2) = 1.0;

    svlGaussian gaussian(mu, sigma);

    cout << "Mean:\n" << toString(*gaussian.mean()) << endl;
    cout << "Variance:\n" << toString(*gaussian.covariance()) << endl;

    cvReleaseMat(&sigma);
    cvReleaseMat(&mu);

    vector<double> x(2);
    vector<int> indx(2);

    x[0] = 1.9; indx[0] = 1;
    x[1] = 3.2; indx[1] = 2;

    svlGaussian conditionalGaussian = gaussian.reduce(x, indx);

    cout << "Mean:\n" << toString(*conditionalGaussian.mean()) << endl;
    cout << "Variance:\n" << toString(*conditionalGaussian.covariance()) << endl;
}

void testSufficientStatistics()
{
    // data
    double x[10][3] = {
        {0.7060, 0.4387, 0.2760},
        {0.0318, 0.3816, 0.6797},
        {0.2769, 0.7655, 0.6551},
        {0.0462, 0.7952, 0.1626},
        {0.0971, 0.1869, 0.1190},
        {0.8235, 0.4898, 0.4984},
        {0.6948, 0.4456, 0.9597},
        {0.3171, 0.6463, 0.3404},
        {0.9502, 0.7094, 0.5853},
        {0.0344, 0.7547, 0.2238}};

    // expected output
    double mu[3] = {0.3978, 0.5614, 0.4500};
    double sigma[3][3] = {
        {0.1171, -0.0027, 0.0376},
        {-0.0027,  0.0369, -0.0017},
        {0.0376, -0.0017, 0.0661}};     // normalized by n

    // accumulate sufficient statistics
    svlSufficientStats stats(3);
    for (int i = 0; i < 10; i++) {
        vector<double> x_i(3);
        for (int j = 0; j < 3; j++) {
            x_i[j] = x[i][j];
        }
        
        stats.accumulate(x_i);
    }

    // convert to gaussian
    svlGaussian g(stats);
    
    cout << "Mean:\n" << toString(*g.mean()) << endl;
    cout << "Covariance:\n" << toString(*g.covariance()) << endl;
}
