/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    buildImageClassificationCache.cpp
** AUTHOR(S):   Olga Russakovsky <olga@cs.stanford.edu>
**              
** DESCRIPTION:
**  Application for caching the results of classifying images, for fast
**  evaluation later.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <string.h>

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

using namespace std;

#define WINDOW_NAME "classifyImages"

// main -----------------------------------------------------------------------

void usage()
{
  //assert(false);
    cerr << SVL_USAGE_HEADER << endl;
    cerr << "USAGE: ./buildImageClassificationCache [OPTIONS] [<dictionary>] <model> (<image>* | <imageSequence>)" << endl;
    cerr << "OPTIONS:" << endl
	 << "  -c <type>             :: classifier type: GABOR, PATCH (default)" << endl
	 << "  -n <object name>      :: name of object" << endl
	 << "  -o <directory>        :: output to directory" << endl
	 << "  -profile              :: turn on SVL code profiling" << endl
	 << "  -s <w> <h>            :: resize image to <w>-by-<h> before scanning" << endl
	 << "  -t <threshold>        :: classifier threshold (default: 0.01)" << endl
         << "  -v                    :: verbose" << endl
         << "  -baseExt              :: sets the file extension that should be stripped off of the input filenames in order to change channels" << endl
         << "  -ch <type> <ext>      :: adds a channel" << endl
	 << "DEPRECATED OPTIONS: " << endl
	 << "  -ch n <string1>..<stringn> :: n input channels, along with types for each of them"<< endl
	 << "  -imageExt <string> :: extension of the image data"<< endl
	 << "  -depthMapExt <string> :: extension of the depth map data if any"<< endl
	 << endl;
}



int main(int argc, char *argv[])
{		
    svlLogger::setLogLevel(SVL_LOG_MESSAGE);
    
    // read commandline parameters
    const char *dictionaryFilename = NULL;
    const char *classifierFilename = NULL;
    const char *outputDirectory = NULL;
    const char *objectName = "[unknown]";
    const char *classifierType = "patch";
    double threshold = 0.01;
    int width = -1;
    int height = -1;

    svlPatchUtil util(SVL_PATCH_CLI_BASE_EXT | SVL_PATCH_DEFAULT_EDGE);    

    char **args = argv + 1;
    while (--argc) {
      if (util.parseCommandLineItem(argc, args))
	{
	  //Do nothing, util has handled this flag 
	}
      else if (!strcmp(*args, "-n")) {
	objectName = *(++args);              argc--;
      } else if (!strcmp(*args, "-c")) {	    
	classifierType = *(++args);          argc--;
      } else if (!strcmp(*args, "-o")) {
	outputDirectory = *(++args);          argc--;
      } else if (!strcmp(*args, "-profile")) {
	svlCodeProfiler::enabled = true;
	svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));
      } else if (!strcmp(*args, "-s")) {
	width = atoi(*(++args));             argc--;
	height = atoi(*(++args));            argc--;
      } else if (!strcmp(*args, "-t")) {
	threshold = atof(*(++args));         argc--;
      } else if (!strcmp(*args, "-v")) {
	svlLogger::setLogLevel(SVL_LOG_VERBOSE);
      } else if ((*args)[0] == '-') {
	SVL_LOG(SVL_LOG_FATAL, "unrecognized option " << *args);
      } else {
	break;
      }
      args++;
    }

    if (argc < (!strcasecmp(classifierType, "patch") ? 3 : 2)) {
	usage();
	return -1;
    }

    if (!strcasecmp(classifierType, "patch")) {
	dictionaryFilename = *args++; argc--;
    }
    classifierFilename = *args++; argc--;


    util.finishCommandLine();

    util.setResizeSize(width, height);
    
    // process image names
    vector<string> imageNames;
    for (int index = 0; index < argc; index++) {
      string imageFilename = string(args[index]);
      if (strExtension(imageFilename).compare("xml") == 0) {
	svlImageSequence sequence;
	sequence.load(imageFilename.c_str());
	for (unsigned i = 0; i < sequence.size(); i++) {
	  imageNames.push_back(sequence.directory() + string("/") + sequence[i]);
	  
	}
      } else {
	imageNames.push_back(imageFilename);
      }
    }
    
    // initialize object detector
    svlSlidingWindowDetector *classifier = NULL;
    if (!strcasecmp(classifierType, "patch")) {
      classifier = new svlPatchBasedObjectDetector(objectName);
      SVL_LOG(SVL_LOG_VERBOSE, "Loading dictionary " << dictionaryFilename << "...");
      ((svlPatchBasedObjectDetector *)classifier)->readDictionary(dictionaryFilename);
    }
    assert(classifier != NULL);
    SVL_LOG(SVL_LOG_VERBOSE, "Loading classifier " << classifierFilename << "...");
    classifier->readModel(classifierFilename);
    classifier->setThreshold(threshold);
    // run a (very slow) detailed scan
    classifier->setDeltas(1, 1, 1.1);

    // run classifier on all supplied images

    svlSoftEdgeMap softEdgeCalculator;

    SVL_LOG(SVL_LOG_MESSAGE, "Running classifier on " << imageNames.size() << " images/sequences...");
    
    for (unsigned index = 0; index < imageNames.size(); index++) {
      //TODO-- share code with buildPatchDictionary and buildPatchResponseCache
      //TODO-- new implementation does not prevent repeated loading of same file. restore that feature
      
      SVL_LOG(SVL_LOG_VERBOSE, "Processing " << imageNames[index] << "...");
      
      string baseName = imageNames[index];
      
      vector<IplImage *> images;
      
      bool success = util.getFrameFromBasename(baseName, images);
      
      if (!success)
	continue;
	
      // classify image, at very detailed scale
      svlObject2dFrame objects;
      objects = classifier->classifyImage(images);
      
      // output objects
      if (outputDirectory) {
	writeCacheObject2dFrame(outputDirectory,
				strBaseName(imageNames[index]).c_str(),
				objects);
      }
      
      // free image
      for (unsigned i = 0; i < images.size(); i++) {
	cvReleaseImage(&images[i]);
      }
    }//image names in image sequence loop
    
    // free classifier and results
    delete classifier;
    
    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);

    return 0;
}
