/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    trainObjectDetector.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**   Trains an object classifier using OpenCV's machine learning library.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <sys/types.h>
#include <vector>

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#include "win32/dirent.h"
#else
#include <dirent.h>
#endif

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"
#include "svlDevel.h"

using namespace std;

// Main -----------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << endl;
    cerr << "USAGE: ./trainObjectDetector [OPTIONS] <+ cache dir> <- cache dir>" << endl;
    cerr << "OPTIONS:" << endl
	 << "  -c <type>           :: classifier type: GENERAL (default), PATCH" << endl
         << "  -i [<dict>] <model> :: evaluate input classifier model (for debugging)" << endl
	 << "  -o <filename>       :: output classifier model" << endl
	 << "  -maxInstances <n>   :: maximum training images for each class (default: 1000)" << endl
	 << "  -rounds <n>         :: number of rounds of boosting (default: 50)" << endl
	 << "  -splits <n>         :: number of splits in weak learners (default: 2)" << endl
         << "  -v                  :: verbose" << endl
         << "  -cv <+ cache dir> <- cache dir>  :: cross-validation examples" << endl
	 << "  -noretrain          :: doesn't re-train on full train+cv set" << endl
	 << "  -cvn <n>            :: performs n-fold cv to find optimal number of boosting rounds" << endl
	 << "  -shuffle            :: shuffles the training set (useful to make sure -cvn is a random subset)" << endl
	 << "  -binary             :: instead of the +/- cache directories just takes the" << endl
	 << "                         names of the binary files" << endl
         << "  -f                  :: uses F-score, rather than normed F-score" 
	 << endl;
}

int main(int argc, char *argv[])
{

  // commented out to get reproduceable results for now
  //srand(time(NULL));

    // read commandline parameters
    const int NUM_REQUIRED_PARAMETERS = 2;

    const char *posCacheDir = NULL;
    const char *negCacheDir = NULL;

    const char *posCVcacheDir = NULL;
    const char *negCVcacheDir = NULL;
    int kFold = -1;

    const char *classifierType = "general";
    const char *inputFilename = NULL;
    const char *dictionaryFilename = NULL;
    const char *outputFilename = NULL;    
    int boostingRounds = 50;
    int weakLearnerSplits = 2;
    int maxTrainingInstances = 1000;

    bool retrain = true;
    bool shuffle = false;
    bool binary = false;

    bool useNormedFScore = true;

    char **args = argv + 1;
    while (--argc > NUM_REQUIRED_PARAMETERS) {
        if (!strcmp(*args, "-c")) {
            classifierType = *(++args); argc--;
        } else if (!strcmp(*args, "-i")) {
	  if (!strcasecmp(classifierType, "patch") || !strcasecmp(classifierType, "general")) {
                dictionaryFilename = *(++args); argc--;
            }
            inputFilename = *(++args); argc--;
        } else if (!strcmp(*args, "-o")) {
            outputFilename = *(++args); argc--;
        } else if (!strcmp(*args, "-rounds")) {
            boostingRounds = atoi(*(++args)); argc--;
        } else if (!strcmp(*args, "-splits")) {
            weakLearnerSplits = atoi(*(++args)); argc--;
        } else if (!strcmp(*args, "-maxInstances")) {
            maxTrainingInstances = atoi(*(++args)); argc--;
        } else if (!strcmp(*args, "-v")) {
            svlLogger::setLogLevel(SVL_LOG_VERBOSE);
	} else if (!strcmp(*args, "-cv")) {
            posCVcacheDir = *(++args); argc--;
            negCVcacheDir = *(++args); argc--;
	} else if (!strcmp(*args, "-cvn")) {
	  kFold = atoi(*(++args)); argc--;
        } else if (!strcmp(*args, "-noretrain")) {
	  retrain = false;        
	} else if (!strcmp(*args, "-shuffle")) {
	  shuffle = true;
	} else if (!strcmp(*args, "-binary")) {
	  binary = true;
	} else if (!strcmp(*args, "-f")) {
	  useNormedFScore = false;
	  cout << "Disabled normed F score" << endl;
	} else {
            SVL_LOG(SVL_LOG_FATAL, "unrecognized option " << *args);
        }
        args++;
    }

    if (argc != NUM_REQUIRED_PARAMETERS) {
        usage();
        return -1;
    }



    posCacheDir = args[0];
    negCacheDir = args[1];

    // define classifier type
    svlSlidingWindowDetector *classifier = NULL;
    if (!strcasecmp(classifierType, "patch")) 
      {
	cout << "Created patch based object detector" << endl;
        classifier = new svlPatchBasedObjectDetector();
      } 
    else if (!strcasecmp(classifierType, "general"))
      {
	cout << "Created general object detector" << endl;
	classifier = new svlGeneralObjectDetector();
      }
    else 
      {
        SVL_LOG(SVL_LOG_FATAL, "classifier type " << classifierType << " not supported");
      }
    assert(classifier != NULL);
    
    // load positive and negative training samples from cache
    SVL_LOG(SVL_LOG_VERBOSE, "Reading positive samples...");
    vector<vector<double> > posSamples = svlReadExamplesFromCache(posCacheDir, maxTrainingInstances, binary);
    SVL_LOG(SVL_LOG_VERBOSE, "..." << posSamples.size() << " samples read");
    SVL_LOG(SVL_LOG_VERBOSE, "Reading negative samples...");
    vector<vector<double> > negSamples = svlReadExamplesFromCache(negCacheDir, maxTrainingInstances, binary);
    SVL_LOG(SVL_LOG_VERBOSE, "..." << negSamples.size() << " samples read");

    // load positive and negative cv samples from cache (unlimited in number)
    vector<vector<double> > posCVsamples;
    vector<vector<double> > negCVsamples;
    if (posCVcacheDir) {
      SVL_LOG(SVL_LOG_VERBOSE, "Reading positive cv samples...");
      posCVsamples = svlReadExamplesFromCache(posCVcacheDir, INT_MAX, binary);
      SVL_LOG(SVL_LOG_VERBOSE, "..." << posCVsamples.size() << " samples read");
    }
    if (negCVcacheDir) {
      SVL_LOG(SVL_LOG_VERBOSE, "Reading negative cv samples...");
      negCVsamples = svlReadExamplesFromCache(negCVcacheDir, INT_MAX, binary);
      SVL_LOG(SVL_LOG_VERBOSE, "..." << negCVsamples.size() << " samples read");
    }

    if (shuffle) {
      random_shuffle(negSamples.begin(), negSamples.end());
      random_shuffle(posSamples.begin(), posSamples.end());
      random_shuffle(negCVsamples.begin(), negCVsamples.end());
      random_shuffle(posCVsamples.begin(), posCVsamples.end());
    }

    // evaluate model
    if (inputFilename != NULL)
    {
        classifier->readModel(inputFilename);
        if (!strcasecmp(classifierType, "patch")) 
        {
            assert(dictionaryFilename != NULL);
            ((svlPatchBasedObjectDetector *)classifier)->readDictionary(dictionaryFilename);
        }
	else if (!strcasecmp(classifierType, "general"))
	  {
	    assert(dictionaryFilename != NULL);
	    svlGeneralObjectDetector * gen = dynamic_cast<svlGeneralObjectDetector *>(classifier);
	    assert(gen);
	    gen->readFeatures(dictionaryFilename);
	  }
  
	classifier->modelPerformance(posSamples, negSamples, "", false);

        if (outputFilename == NULL) {
            delete classifier;
            return 0;
        }
    }

    // set classifier learning options
    classifier->setOption("boostingRounds", boostingRounds);
    classifier->setOption("numSplits", weakLearnerSplits);
    classifier->setOption("maxTrainingImages", maxTrainingInstances);

    SVL_LOG(SVL_LOG_VERBOSE, "Training classifier...");

    if (posCVsamples.size() > 0 && negCVsamples.size() > 0) {
      classifier->learnModelCV(posSamples, negSamples, posCVsamples, negCVsamples, retrain, useNormedFScore);
    } else if (kFold > 0) {
      classifier->learnModelFullCV(posSamples, negSamples, kFold, useNormedFScore);
    } else {
      classifier->learnModel(posSamples, negSamples, useNormedFScore);
    }

    if (outputFilename != NULL) {
        SVL_LOG(SVL_LOG_VERBOSE, "Writing model to file " << outputFilename);
        classifier->writeModel(outputFilename);
    }


    // free memory
    delete classifier;

    return 0;
}

