/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaCommon.h
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Asdf ...
**
*****************************************************************************/

#pragma once

#ifndef __SVL_CUDA_COMMON_H
#define __SVL_CUDA_COMMON_H

#define USE_CUDA

#ifdef USE_CUDA
#ifdef _WIN32
#	define WINDOWS_LEAN_AND_MEAN
#	include <windows.h>
//#	include <C:\Program Files\NVIDIA Corporation\NVIDIA CUDA SDK\common\inc\cuda.h>
//#	include <C:\Program Files\NVIDIA Corporation\NVIDIA CUDA SDK\common\inc\cuda_gl_interop.h>
#	include <C:\Program Files\NVIDIA Corporation\NVIDIA CUDA SDK\common\inc\cutil.h>
#else
#	include <cuda.h>
#	include <cuda_gl_interop.h>
#	include <cutil.h>
#endif
#endif

#include <iostream>
#include <vector>

#include "cv.h"
#include "cxcore.h"
using namespace std;

extern int iDivUp(int a, int b);
extern int iDivDown(int a, int b);
extern int iAlignUp(int a, int b);
extern int iAlignDown(int a, int b);

// Struct for controlling 2D memory on the device (GPU).
struct svlCudaPitch {
	int pitch, w, h;	// Device pitch, width, and height.
	float *data;		// Device pointer
	void *source;		// Host data pointers
	svlCudaPitch() : pitch(-1), w(-1), h(-1) {
		data = NULL;
		source = NULL;
	}
	inline int size() { return w*h; }
};

// Struct for controlling 1D memory on the device.
struct svlCudaLine {
	int w;				// Width
	float *data;		// Device pointer
	void *source;		// Host data pointer
	svlCudaLine() : w(-1) {
		data = NULL;
		source = NULL;
	}
};

// Struct for binary masks on which blocks in a cuda function need to execute.
struct svlCudaCoveringGrid {
	int grid_x, grid_y;	// grid x and y sizes of grid.
	float *data;		// The grid data

	// Initialize with grid sizes directly.
	inline svlCudaCoveringGrid(int grid_x, int grid_y)
		: grid_x(grid_x), grid_y(grid_y) {
			data = (float*)malloc(sizeof(float)*grid_x*grid_y);
	}
	inline svlCudaCoveringGrid(int grid_x, int grid_y, float *data)
		: grid_x(grid_x), grid_y(grid_y), data(data) {}
	// Initialize with image and block sizes.
	inline svlCudaCoveringGrid(int img_w, int img_h, int block_x, int block_y) {
		grid_x = iDivUp(img_w, block_x);
		grid_y = iDivUp(img_h, block_y);
		data = (float*)malloc(sizeof(float)*grid_x*grid_y);
	}
	~svlCudaCoveringGrid() {
		if ( data != NULL ) free(data);
	}

	// Other members
	inline int size() {
		return grid_x * grid_y;
	}
	inline void clearGrid() {
		memset(data, 0, grid_x*grid_y);
	}
	// Direct element access functions.
	inline float operator[](int i) const {
		return this->data[i];
	}
	inline float& operator[](int i) {
		return this->data[i];
	}
};

// Cuda functions
extern "C" {
	// *** svlCudaCommon.cu functions ***
	// Pitch functions
	void svlCudaAllocPitch(int w, int h, svlCudaPitch &pitch);
	void svlCudaAllocPitches(int w, int h, vector<svlCudaPitch*> &pitches, int n = 1);
	void svlCudaFreePitch(svlCudaPitch &p);
	void svlCudaFreePitches(vector<svlCudaPitch> &p);
	void svlCudaMemset2D(svlCudaPitch &p, int value = 0);
	void svlCudaPitchToGPU(svlCudaPitch &p, const float *src, int w = -1, int h = -1);
	void svlCudaPitchToGPUasync(svlCudaPitch &p, const float *src, int w = -1, int h = -1);
	void svlCudaPitchToGPU_all(svlCudaPitch &p, const float *src, int d_pitch, int w = -1, int h = -1);
	void svlCudaPitchFromGPU(float *dest, const svlCudaPitch &p, int w = -1, int h = -1);
	void svlCudaPitchFromGPU_all(float *dest, int d_pitch, int w, int h, const svlCudaPitch &p);
	void svlCudaPitchToPitch(svlCudaPitch &dest, svlCudaPitch &src, int w = -1, int h = -1);
	// Line functions
	void svlCudaAllocLine(int w, svlCudaLine &n);
	void svlCudaFreeLine(svlCudaLine &n);
	void svlCudaLineToGPU(svlCudaLine &n, const float *src, int w = -1);
	void svlCudaLineFromGPU(float *dest, const svlCudaLine &n, int w = -1);
	// Other accessing functions
	void svlCudaRowRandomAccess(const IplImage *ixes, IplImage *data_out, const vector<int> row_counts, float *dest, svlCudaPitch &data, svlCudaPitch &addys);
	void svlCudaLinearRandomAccess(const IplImage *ixes, int max_row, float *dest, svlCudaPitch &data, svlCudaPitch &addys, svlCudaPitch &data_out);
	void svlCudaLine2DRandomAccess(svlCudaLine *ixline, svlCudaPitch *data, svlCudaLine *res, int res_off, int res_stride, int res_n);
	void svlCudaAddCvPoint2Line(svlCudaLine *lin, CvPoint &pt, svlCudaLine *lout, int w = -1);

	// *** svlCudaConvolution.cu functions ***
	// Plain convolution functions.
	void svlCudaConvolve2d(const svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, float *patch, int patch_w, int patch_h, int img_w = -1, int img_h = -1);
	void svlCudaConvolve2d_v2(const svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, float *patch, int patch_w, int patch_h, int img_w = -1, int img_h = -1);
	void svlCudaConvolve2d_v3(const svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, float *patch, int patch_w, int patch_h, int img_w = -1, int img_h = -1);
	// Normalized crosscorrelation functions.
	void svlCudaNormCCorr2d(const svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, float *patch, float *mask,
		int patch_w, int patch_h, int patch_real_numel, float patch_mean = 0.0f, float patch_std = 1.0f, int img_w = -1, int img_h = -1,
		svlCudaCoveringGrid *cudagrid = NULL);
	void svlCudaNormCCorr2d_v2(const svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, float *patch, float *mask,
		int patch_w, int patch_h, int patch_real_numel, float patch_mean = 0.0f, float patch_std = 1.0f, int img_w = -1, int img_h = -1, bool cached = false);
	void svlCudaNormCCorr2d_twostep(const svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, svlCudaPitch &temp1, svlCudaPitch &temp2,
		float *patch, float *mask, int patch_w, int patch_h, int patch_real_numel,
		float patch_mean, float patch_std, int img_w, int img_h);
	void svlCudaNormCCorr2d_v3(const svlCudaPitch *pin, svlCudaPitch *pout, float *patch, float *mask, svlCudaPitch *integ, svlCudaPitch *integ2,
		int patch_w, int patch_h, int patch_real_numel, float patch_mean, float patch_std, int img_w, int img_h, bool cached = false);

	// *** svlCudaFilter.cu function ***
	void svlCudaSobel(svlCudaPitch *pin, svlCudaPitch *edge, svlCudaPitch *temp1, svlCudaPitch *temp2, int img_w = -1, int img_h = -1);

	// *** svlCudaNormalize.cu functions ***
	void svlCudaWindowNormalize(svlCudaPitch &pitch_in, const svlCudaPitch &integral, const svlCudaPitch &integral2,
		float patch_mean, float patch_std, int patch_w, int patch_h, int img_w = -1, int img_h = -1);
	void svlCudaWindowNormalize1(svlCudaPitch &pitch_in, const svlCudaPitch &integral, const svlCudaPitch &integral2,
		float patch_mean, float patch_std, int patch_w, int patch_h, int img_w = -1, int img_h = -1);
	void svlCudaWindowNormalize1DimSub(svlCudaPitch &pitch_in, const svlCudaPitch &integral, const svlCudaPitch &integral2,
		float patch_mean, float patch_std, int patch_w, int patch_h, int img_w = -1, int img_h = -1, int dim = 128);
	void svlCudaWindowSumSumSq(const svlCudaPitch *pin, svlCudaPitch *sum, svlCudaPitch *sum2, int win_w, int win_h, int img_w = -1, int img_h = -1);
	void svlCudaPatchNormalize(svlCudaPitch *pout, svlCudaPitch *sum, svlCudaPitch *sum2, float patch_mean, float patch_std, int patch_N, int img_w = -1, int img_h = -1);
	void svlCudaWindowNormalizeRow(svlCudaPitch *pin, svlCudaPitch *pout, int patch_w, int img_w = -1, int img_h = -1, int dim = 128);

	// *** svlCudaMath.cu functions ***
	void svlCudaSubtract(svlCudaPitch &a, svlCudaPitch &b, int img_w = -1, int img_h = -1);
	void svlCudaDecimate(svlCudaPitch &a, svlCudaPitch &b, int delta_x = 2, int delta_y = 2, int img_w = -1, int img_h = -1);
	void svlCudaAddQuadrature(svlCudaPitch *a, svlCudaPitch *b, int img_w = -1, int img_h = -1);
	// Min-max'ing
	void svlCudaMax(svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, int max_w = 7, int max_h = 7, int img_w = -1, int img_h = -1, svlCudaCoveringGrid *cudagrid = NULL);
	void svlCudaMax7x7(svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, int img_w = -1, int img_h = -1, svlCudaCoveringGrid *cudagrid = NULL);
	void svlCudaMax7x7Cent(svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, int img_w = -1, int img_h = -1);
	void svlCudaMinMax4x4(svlCudaPitch &pitch_in, svlCudaPitch &min_out, svlCudaPitch &max_out, int img_w, int img_h);
	void svlCudaMinMax32x32(svlCudaPitch &pitch_in, svlCudaPitch &min_out, svlCudaPitch &max_out, int img_w, int img_h);
	// Image resizing
	void svlCudaResize2D(svlCudaPitch &pin, svlCudaPitch &pout, float scale = 0.8f, int img_w = -1, int img_h = -1);

	// *** svlCudaIntegration.cu functions ***
	void svlCudaIntegralLine(svlCudaLine *n, svlCudaLine *integ, int w = -1, int dim = 128, bool do_cumsum = false, int stride = -1);
	//void svlCudaIntegral(svlCudaPitch &pin, svlCudaPitch *integ, svlCudaPitch *integ2 = NULL, int img_w = -1, int img_h = -1, int dim = 128);
	void svlCudaIntegralBlock(svlCudaPitch &pin, svlCudaPitch *integ, svlCudaPitch *tempint, svlCudaPitch *integ2 = NULL, svlCudaPitch *tempint2 = NULL, int img_w = -1, int img_h = -1, int dim = 128);
	void svlCudaRowIntegralBlock(svlCudaPitch &pin, svlCudaPitch *integ, svlCudaPitch *integ2 = NULL, int img_w = -1, int img_h = -1, int dim = 128);

	// *** svlCudaDTree.cu functions ***
	void svlCudaDTreeEvaluate(svlCudaLine &dtree, int n_trees, svlCudaLine &features, int feature_len, int feature_stride, int n_fvs, int execs_per_block);
	void svlCudaDTreeCollectResults(svlCudaLine *features, svlCudaLine *res, int n_trees, int n_fvs, int feature_stride);
};

// Fast integer multiplication and easy pitch accessor
#define IMUL(a, b) __mul24(a, b)
// Takes the data pointer, pitch size (in bytes), and row offset.
#define PITCH_GET(d,p,s) (float*)((char*)d + IMUL(p,s))
#define INT_PITCH_GET(d,p,s) (int*)((char*)d + IMUL(p,s))

#endif

