/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaConvolution.cu
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Cuda convolution and related functions.
**
*****************************************************************************/

#include "svlCudaCommon.h"
#include <limits>

int iDivUp(int a, int b){
	return (a % b != 0) ? (a / b + 1) : (a / b);
}
//Round a/b to nearest lower integer value
int iDivDown(int a, int b){
	return a / b;
}
//Align a to nearest higher multiple of b
int iAlignUp(int a, int b){
	return (a % b != 0) ?  (a - a % b + b) : a;
}
//Align a to nearest lower multiple of b
int iAlignDown(int a, int b){
	return a - a % b;
}

extern bool sizesMatch(svlCudaCoveringGrid *cudagrid, dim3 &grid);

// Device memory.
__device__ __constant__ float kern_patch[256];	// Patch kernel for convolution.
__device__ __constant__ float mask_patch[256];	// Mask kernel for convolution.
// Elsewhere-defined device memory;
//extern __device__ __constant__ float *covering_grid; // For holding bits on whether to operation on this block or not.
__device__ __constant__ float covering_grid[ 1536 ]; // For holding bits on whether to operation on this block or not.

// 2D patch width-by-height convolution. Assumed block size is aligned with image size (so there are no off-image threads).
template<int NUM_COLS, int NUM_ROWS> __global__ void conv2d_WxH_slice(float *din, int din_pitch, float *res, int res_pitch, int data_w, int data_h)
{
	// Load din data into shared cache.
	__shared__ float data[ 1536 ];
	const int data_width = 32;
	const int data_size = IMUL(data_width, blockDim.y);
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.

	// Read in data.
	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;		// Shared thread memory index.
	const float *my_din = PITCH_GET(din, din_pitch, y);				// Calculate my row of din.
	data[ix] = my_din[x];
	// Read in right-ward entry.
	const bool right_on = x + blockDim.x < data_w;
	data[ix + blockDim.x] = right_on ? my_din[x + blockDim.x] : 0.0f;
	// Read in downward entries.
	if ( threadIdx.y < NUM_ROWS && y + blockDim.y < data_h ) {
		// Read in upper entries
		const float *in_up = PITCH_GET(din, din_pitch, y + blockDim.y);
		data[ ix + data_size ] = in_up[x];
		// Read in right-ward entry.
		data[ ix + data_size + blockDim.x ] = right_on ? in_up[x + blockDim.x] : 0.0f;
	} else {
		data[ ix + data_size ] = data[ ix + data_size + blockDim.x ] = 0.0f;
	}

	__syncthreads();

	// Compute my convolution result.
	float gf = 0.0f;
#pragma unroll
	for ( int i=0; i<NUM_ROWS; ++i ) {
#pragma unroll
		for ( int j=0; j<NUM_COLS; ++j ) {
			gf += kern_patch[ IMUL(NUM_COLS,i) + j ] * data[ ix + IMUL(data_width,i) + j ];
		}
	}
	// Calculate my row of the results.
	float *my_res = PITCH_GET(res, res_pitch, y);
	my_res[x] = gf;
}

// Execute a 2D convolution between pitches using a patch.
void svlCudaConvolve2d(const svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, float *patch, int patch_w, int patch_h, int img_w, int img_h)
{
	// Copy patch over to GPU.
	CUDA_SAFE_CALL( cudaMemcpyToSymbol( kern_patch, patch, patch_w*patch_h*sizeof(float), 0, cudaMemcpyHostToDevice) );

	// Convolve.
	dim3 block(16, 16);
	//dim3 grid( iDivUp(pitch_in.w, block.x), iDivUp(pitch_in.h, block.y) );
	dim3 grid( iDivUp( ( img_w == -1 ? pitch_in.w : img_w ), block.x), iDivUp( ( img_h == -1 ? pitch_in.h : img_h ), block.y) );
	//printf("svlCudaConvolve2d :: got (%d,%d) => (%d,%d)\n", img_w, img_h, grid.x, grid.y);

	if ( patch_w <= 4 && patch_h <= 4 ) {
		conv2d_WxH_slice<4,4><<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch,  pitch_in.w, pitch_in.h);
	} else if ( patch_w <= 8 && patch_h <= 8 ) {
		conv2d_WxH_slice<8,8><<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch,  pitch_in.w, pitch_in.h);
	} else if ( patch_w <= 8 && patch_h <= 16 ) {
		conv2d_WxH_slice<8,16><<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch,  pitch_in.w, pitch_in.h);
	} else if ( patch_w <= 16 && patch_h <= 8 ) {
		conv2d_WxH_slice<16,8><<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch,  pitch_in.w, pitch_in.h);
	} else if ( patch_w <= 16 && patch_h <= 16 ) {
		conv2d_WxH_slice<16,16><<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch,  pitch_in.w, pitch_in.h);
	}
}

// 2D patch width-by-height convolution, half-templated on width. Assumed block size is aligned with image size (so there are no off-image threads).
template<int NUM_COLS> __global__ void conv2d_halfT_slice(float *din, int din_pitch, float *res, int res_pitch, int data_w, int data_h, int NUM_ROWS)
{
	// Load din data into shared cache.
	__shared__ float data[ 1536 ];
	const int data_width = 32;
	const int data_size = IMUL(data_width, blockDim.y);
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.

	// Read in data.
	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;		// Shared thread memory index.
	const float *my_din = PITCH_GET(din, din_pitch, y);				// Calculate my row of din.
	data[ix] = my_din[x];
	// Read in right-ward entry.
	const bool right_on = x + blockDim.x < data_w;
	data[ix + blockDim.x] = right_on ? my_din[x + blockDim.x] : 0.0f;
	// Read in downward entries.
	if ( threadIdx.y < NUM_ROWS && y + blockDim.y < data_h ) {
		// Read in upper entries
		const float *in_up = PITCH_GET(din, din_pitch, y + blockDim.y);
		data[ ix + data_size ] = in_up[x];
		// Read in right-ward entry.
		data[ ix + data_size + blockDim.x ] = right_on ? in_up[x + blockDim.x] : 0.0f;
	} else {
		data[ ix + data_size ] = data[ ix + data_size + blockDim.x ] = 0.0f;
	}

	__syncthreads();

	// Compute my convolution result.
	float gf = 0.0f;
#pragma unroll
	for ( int i=0; i<NUM_ROWS; ++i ) {
#pragma unroll
		for ( int j=0; j<NUM_COLS; ++j ) {
			gf += kern_patch[ IMUL(NUM_COLS,i) + j ] * data[ ix + IMUL(data_width,i) + j ];
		}
	}
	// Calculate my row of the results.
	float *my_res = PITCH_GET(res, res_pitch, y);
	my_res[x] = gf;
}

// Execute a 2D convolution between pitches using a patch.
void svlCudaConvolve2d_v2(const svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, float *patch, int patch_w, int patch_h, int img_w, int img_h)
{
	// Copy patch over to GPU.
	CUDA_SAFE_CALL( cudaMemcpyToSymbol( kern_patch, patch, patch_w*patch_h*sizeof(float), 0, cudaMemcpyHostToDevice) );

	// Convolve.
	dim3 block(16, 16);
	//dim3 grid( iDivUp(pitch_in.w, block.x), iDivUp(pitch_in.h, block.y) );
	dim3 grid( iDivUp( ( img_w == -1 ? pitch_in.w : img_w ), block.x), iDivUp( ( img_h == -1 ? pitch_in.h : img_h ), block.y) );
	//printf("svlCudaConvolve2d :: got (%d,%d) => (%d,%d)\n", img_w, img_h, grid.x, grid.y);

	if ( patch_w == 4 ) {
		conv2d_halfT_slice <4> <<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, pitch_in.w, pitch_in.h, patch_h);
	} else if ( patch_w == 8 ) {
		conv2d_halfT_slice <8> <<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, pitch_in.w, pitch_in.h, patch_h);
	} else if ( patch_w == 16 ) {
		conv2d_halfT_slice <16> <<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, pitch_in.w, pitch_in.h, patch_h);
	}
}

// 2D patch width-by-height normalized cross-correlation.
template<int NUM_COLS, int NUM_ROWS>
__device__ void conv2d_WxH_normed_slice(float *din, int din_pitch, float *res, int res_pitch,
										int data_w, int data_h, float patch_mean, float patch_inv_std, float N_inv)
{
	// Load din data into shared cache.
	__shared__ float data[ 1536 ];
	const int data_width = 32;
	const int data_size = IMUL(data_width, blockDim.y);
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.

	// Read in data.
	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;		// Shared thread memory index.
	const float *my_din = PITCH_GET(din, din_pitch, y);				// Calculate my row of din.
	data[ix] = my_din[x];
	// Read in rightward entry.
	const bool right_on = x + blockDim.x < data_w;
	data[ix + blockDim.x] = right_on ? my_din[x + blockDim.x] : 0.0f;
	if ( threadIdx.y < NUM_ROWS ) {
		if ( y + blockDim.y < data_h ) {
			// Read in upper entries
			const float *in_up = PITCH_GET(din, din_pitch, y + blockDim.y);
			data[ ix + data_size ] = in_up[x];
			// Read in rightward entry.
			data[ ix + data_size + blockDim.x ] = right_on ? in_up[x + blockDim.x] : 0.0f;
		} else {
			data[ ix + data_size ] = data[ ix + data_size + blockDim.x ] = 0.0f;
		}
	}

	__syncthreads();

	// Compute my convolution result.
	float gf = 0.0f, window_sum = 0.0f, window_sq_sum = 0.0f;
#pragma unroll
	for ( int i=0; i<NUM_ROWS; ++i ) {
#pragma unroll
		for ( int j=0; j<NUM_COLS; ++j ) {
			const int pix = IMUL(NUM_COLS,i) + j;
			if ( mask_patch[ pix ] != 0.0f ) {
				const float d = data[ ix + IMUL(data_width,i) + j ];
				gf += kern_patch[ pix ] * d;
				window_sum += d;
				window_sq_sum += d * d;
			}
		}
	}

	// Calculate my row of the results.
	float *my_res = PITCH_GET(res, res_pitch, y);
	const float window_mean = window_sum * N_inv;
	const float window_sq_mean = window_sq_sum * N_inv;
	my_res[x] = __fdividef( gf - patch_mean * window_sum, sqrtf(window_sq_mean - window_mean*window_mean) ) * patch_inv_std * N_inv;
}

// Covering grid versions of the functions above.
template<int NUM_COLS, int NUM_ROWS>
__global__ void conv2d_WxH_normed_NoCG_slice(float *din, int din_pitch, float *res, int res_pitch,
											 int data_w, int data_h, float patch_mean, float patch_inv_std, float N_inv)
{
	conv2d_WxH_normed_slice<NUM_COLS, NUM_ROWS>(din, din_pitch, res, res_pitch, data_w, data_h, patch_mean, patch_inv_std, N_inv);
}
template<int NUM_COLS, int NUM_ROWS>
__global__ void conv2d_WxH_normed_CG_slice(float *din, int din_pitch, float *res, int res_pitch,
										   int data_w, int data_h, float patch_mean, float patch_inv_std, float N_inv)
{
	if ( covering_grid[ IMUL(blockIdx.y, gridDim.x) + blockIdx.x ] != 0.0f ) {
		conv2d_WxH_normed_slice<NUM_COLS, NUM_ROWS>(din, din_pitch, res, res_pitch, data_w, data_h, patch_mean, patch_inv_std, N_inv);
	}
}

// Execute a 2D convolution between pitches using a patch.
void svlCudaNormCCorr2d(const svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, float *patch, float *mask,
						int patch_w, int patch_h, int patch_real_numel, float patch_mean, float patch_std, int img_w, int img_h,
						svlCudaCoveringGrid *cudagrid)
{
	// Copy patches over to GPU.
	CUDA_SAFE_CALL( cudaMemcpyToSymbol( kern_patch, patch, patch_w*patch_h*sizeof(float), 0, cudaMemcpyHostToDevice) );
	CUDA_SAFE_CALL( cudaMemcpyToSymbol( mask_patch, mask, patch_w*patch_h*sizeof(float), 0, cudaMemcpyHostToDevice) );

	// Convolve.
	dim3 block(16, 16);
	dim3 grid( iDivUp( ( img_w == -1 ? pitch_in.w : img_w ), block.x), iDivUp( ( img_h == -1 ? pitch_in.h : img_h ), block.y) );

	// Pre-reciprocalize standard deviation
	patch_std = 1.0f / patch_std;

	if ( cudagrid == NULL || !sizesMatch(cudagrid, grid) ) {
		// Use non-covering grid functions.
		if ( patch_w <= 4 && patch_h <= 4 ) {
			conv2d_WxH_normed_NoCG_slice <4,4> <<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch, pitch_in.w, pitch_in.h,
				patch_mean, patch_std, 1.0f/float(patch_real_numel));
		}
		else if ( patch_w <= 8 && patch_h <= 8 ) {
			conv2d_WxH_normed_NoCG_slice <8,8> <<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch, pitch_in.w, pitch_in.h,
				patch_mean, patch_std, 1.0f/float(patch_real_numel));
		}
		else if ( patch_w <= 8 && patch_h <= 16 ) {
			conv2d_WxH_normed_NoCG_slice <8,16> <<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch, pitch_in.w, pitch_in.h,
				patch_mean, patch_std, 1.0f/float(patch_real_numel));
		}
		else if ( patch_w <= 16 && patch_h <= 8 ) {
			conv2d_WxH_normed_NoCG_slice <16,8> <<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch, pitch_in.w, pitch_in.h,
				patch_mean, patch_std, 1.0f/float(patch_real_numel));
		}
		else if ( patch_w <= 16 && patch_h <= 16 ) {
			conv2d_WxH_normed_NoCG_slice <16,16> <<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch, pitch_in.w, pitch_in.h,
				patch_mean, patch_std, 1.0f/float(patch_real_numel));
		}
	} else {
		// Use covering grid functions.

		// Copy grid to device.
		CUDA_SAFE_CALL( cudaMemcpyToSymbol( covering_grid, cudagrid->data, cudagrid->size()*sizeof(float), 0, cudaMemcpyHostToDevice) );

		if ( patch_w <= 4 && patch_h <= 4 ) {
			conv2d_WxH_normed_CG_slice <4,4> <<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch, pitch_in.w, pitch_in.h,
				patch_mean, patch_std, 1.0f/float(patch_real_numel));
		}
		else if ( patch_w <= 8 && patch_h <= 8 ) {
			conv2d_WxH_normed_CG_slice <8,8> <<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch, pitch_in.w, pitch_in.h,
				patch_mean, patch_std, 1.0f/float(patch_real_numel));
		}
		else if ( patch_w <= 8 && patch_h <= 16 ) {
			conv2d_WxH_normed_CG_slice <8,16> <<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch, pitch_in.w, pitch_in.h,
				patch_mean, patch_std, 1.0f/float(patch_real_numel));
		}
		else if ( patch_w <= 16 && patch_h <= 8 ) {
			conv2d_WxH_normed_CG_slice <16,8> <<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch, pitch_in.w, pitch_in.h,
				patch_mean, patch_std, 1.0f/float(patch_real_numel));
		}
		else if ( patch_w <= 16 && patch_h <= 16 ) {
			conv2d_WxH_normed_CG_slice <16,16> <<< grid, block >>>(pitch_in.data, pitch_in.pitch,  pitch_out.data, pitch_out.pitch, pitch_in.w, pitch_in.h,
				patch_mean, patch_std, 1.0f/float(patch_real_numel));
		}
	}
}

// 2D patch width-by-height normalized cross-correlation.
template<int NUM_COLS>
__global__ void conv2d_halfT_normed_slice(float *din, int din_pitch, float *res, int res_pitch,
										  int data_w, int data_h, float patch_mean, float patch_inv_std, float N_inv, int NUM_ROWS)
{
	// Load din data into shared cache.
	__shared__ float data[ 1024 ];
	const int data_width = 32;
	const int data_size = IMUL(data_width, blockDim.y);
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.

	// Read in data.
	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;		// Shared thread memory index.
	const float *my_din = PITCH_GET(din, din_pitch, y);				// Calculate my row of din.
	data[ix] = my_din[x];
	// Read in rightward entry.
	const bool right_on = x + blockDim.x < data_w;
	data[ix + blockDim.x] = right_on ? my_din[x + blockDim.x] : 0.0f;
	if ( threadIdx.y < NUM_ROWS ) {
		if ( y + blockDim.y < data_h ) {
			// Read in upper entries
			const float *in_up = PITCH_GET(din, din_pitch, y + blockDim.y);
			data[ ix + data_size ] = in_up[x];
			// Read in rightward entry.
			data[ ix + data_size + blockDim.x ] = right_on ? in_up[x + blockDim.x] : 0.0f;
		} else {
			data[ ix + data_size ] = data[ ix + data_size + blockDim.x ] = 0.0f;
		}
	}

	__syncthreads();

	// Compute my convolution result.
	float gf = 0.0f, window_sum = 0.0f, window_sq_sum = 0.0f;
#pragma unroll
	for ( int i=0; i<NUM_ROWS; ++i ) {
#pragma unroll
		for ( int j=0; j<NUM_COLS; ++j ) {
			const int pix = IMUL(NUM_COLS,i) + j;
			if ( mask_patch[ pix ] != 0.0f ) {
				const float d = data[ ix + IMUL(data_width,i) + j ];
				gf += kern_patch[ pix ] * d;
				window_sum += d;
				window_sq_sum += d * d;
			}
		}
	}

	// Calculate my row of the results.
	float *my_res = PITCH_GET(res, res_pitch, y);
	const float window_mean = window_sum * N_inv;
	const float window_sq_mean = window_sq_sum * N_inv;
	my_res[x] = __fdividef( gf - patch_mean * window_sum, sqrtf(window_sq_mean - window_mean*window_mean) ) * patch_inv_std * N_inv;
}

// Normalized cross correlation with half-T convolution.
void svlCudaNormCCorr2d_v2(const svlCudaPitch &pin, svlCudaPitch &pout, float *patch, float *mask,
						   int patch_w, int patch_h, int patch_real_numel, float patch_mean, float patch_std, int img_w, int img_h, bool cached)
{
	// Copy patches over to GPU.
	if ( !cached ) {
		// We were given host pointer for patch and mask.
		CUDA_SAFE_CALL( cudaMemcpyToSymbol( kern_patch, patch, patch_w*patch_h*sizeof(float), 0, cudaMemcpyHostToDevice) );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol( mask_patch, mask, patch_w*patch_h*sizeof(float), 0, cudaMemcpyHostToDevice) );
	} else {
		// Device pointers; yay.
		CUDA_SAFE_CALL( cudaMemcpyToSymbol( kern_patch, patch, patch_w*patch_h*sizeof(float), 0, cudaMemcpyDeviceToDevice) );
		CUDA_SAFE_CALL( cudaMemcpyToSymbol( mask_patch, mask, patch_w*patch_h*sizeof(float), 0, cudaMemcpyDeviceToDevice) );
	}

	// Convolve.
	dim3 block(16, 16);
	if ( img_w == -1 ) img_w = pin.w;
	if ( img_h == -1 ) img_h = pin.h;
	dim3 grid( iDivUp(img_w, block.x), iDivUp(img_h, block.y) );

	patch_std = 1.0f / patch_std; // Pre-reciprocalize standard deviation

	float N_inv = 1.0f / float(patch_real_numel);
	if ( patch_w == 4 ) {
		conv2d_halfT_normed_slice <4> <<< grid, block >>>(pin.data, pin.pitch, pout.data, pout.pitch, img_w, img_h, patch_mean, patch_std, N_inv, patch_h);
	} else if ( patch_w == 8 ) {
		conv2d_halfT_normed_slice <8> <<< grid, block >>>(pin.data, pin.pitch, pout.data, pout.pitch, img_w, img_h, patch_mean, patch_std, N_inv, patch_h);
	} else if ( patch_w == 16 ) {
		conv2d_halfT_normed_slice <16> <<< grid, block >>>(pin.data, pin.pitch, pout.data, pout.pitch, img_w, img_h, patch_mean, patch_std, N_inv, patch_h);
	}
}

// 2D patch width-by-height normalized cross-correlation.
template<int NUM_COLS, int NUM_ROWS> __global__ void crosscorr_multi_slice(float *din, int din_pitch, float *res, int res_pitch,
																		   float *sum_data, int sum_pitch, float *sum_sq_data, int sum_sq_pitch,
																		   int data_w, int data_h, float patch_mean, float patch_inv_std, float N_inv)
{
	// Load din data into shared cache.
	__shared__ float data[ 1536 ];
	const int data_width = 32;
	const int data_size = IMUL(data_width, blockDim.y);
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.

	// Read in data.
	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;		// Shared thread memory index.
	const float *my_din = PITCH_GET(din, din_pitch, y);				// Calculate my row of din.
	data[ix] = my_din[x];
	if ( x + blockDim.x < data_w ) {
		// Read in right-ward entry.
		data[ix + blockDim.x] = my_din[x + blockDim.x];
	} else {
		data[ix + blockDim.x] = 0.0f;
	}
	if ( threadIdx.y < NUM_ROWS && y + blockDim.y < data_h ) {
		// Read in upper entries
		const float *in_up = PITCH_GET(din, din_pitch, y + blockDim.y);
		data[ ix + data_size ] = in_up[x];
		if ( x + blockDim.x < data_w ) {
			// Read in right-ward entry.
			data[ ix + data_size + blockDim.x ] = in_up[x + blockDim.x];
		} else {
			data[ ix + data_size + blockDim.x ] = 0.0f;
		}
	} else {
		data[ ix + data_size ] = 0.0f;
		data[ ix + data_size + blockDim.x ] = 0.0f;
	}

	__syncthreads();

	// Compute my convolution result.
	float gf = 0.0f, window_sum = 0.0f, window_sq_sum = 0.0f, d;
#pragma unroll
	for ( int i=0; i<NUM_ROWS; ++i ) {
#pragma unroll
		for ( int j=0; j<NUM_COLS; ++j ) {
			if ( mask_patch[ IMUL(NUM_COLS,i) + j ] != 0.0f ) {
				d = data[ ix + IMUL(data_width,i) + j ];
				gf += kern_patch[ IMUL(NUM_COLS,i) + j ] * d;
				window_sum += d;
				window_sq_sum += d * d;
			}
		}
	}

	// Calculate my row of the results.
	float *my_res = PITCH_GET(res, res_pitch, y);
	float *sum_out = PITCH_GET(sum_data, sum_pitch, y);
	float *sum_sq_out = PITCH_GET(sum_sq_data, sum_sq_pitch, y);
	my_res[x] = gf;
	sum_out[x] = window_sum;
	sum_sq_out[x] = window_sq_sum;
}

// 2D patch width-by-height normalized cross-correlation.
__global__ void crosscorr_normit_slice(float *res, int res_pitch, float *sum_data, int sum_pitch, float *sum_sq_data, int sum_sq_pitch,
									   int img_w, int img_h, float patch_mean, float patch_inv_std, float N_inv)
{
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.
	if ( x >= img_w || y >= img_h )
		return;

	float *res_it = PITCH_GET(res, res_pitch, y);
	const float *sum = PITCH_GET(sum_data, sum_pitch, y);
	const float *sum_sq = PITCH_GET(sum_sq_data, sum_sq_pitch, y);
	res_it[x] = __fdividef( res_it[x] - patch_mean * sum[x], __powf( N_inv * (sum_sq[x] - sum[x]*sum[x]*N_inv), 0.5f) ) * patch_inv_std * N_inv;
	/*const float subtraction = res_it[x] - patch_mean * sum[x];
	const float denominator = sum_sq[x] - sum[x] * sum[x] / N_inv;
	res_it[x] = subtraction / ( patch_inv_std * sqrt( denominator ) );*/
}


// Execute a 2D convolution between pitches using a patch.
void svlCudaNormCCorr2d_twostep(const svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, svlCudaPitch &temp1, svlCudaPitch &temp2,
								float *patch, float *mask, int patch_w, int patch_h, int patch_real_numel,
								float patch_mean, float patch_std, int img_w, int img_h)
{
	// Copy patches over to GPU.
	CUDA_SAFE_CALL( cudaMemcpyToSymbol( kern_patch, patch, patch_w*patch_h*sizeof(float), 0, cudaMemcpyHostToDevice) );
	CUDA_SAFE_CALL( cudaMemcpyToSymbol( mask_patch, mask, patch_w*patch_h*sizeof(float), 0, cudaMemcpyHostToDevice) );

	// Convolve.
	dim3 block(16, 16);
	dim3 grid( iDivUp( ( img_w == -1 ? pitch_in.w : img_w ), block.x), iDivUp( ( img_h == -1 ? pitch_in.h : img_h ), block.y) );

	// Pre-reciprocalize standard deviation
	patch_std = 1.0f / patch_std; // Invert standard deviation.
	if ( patch_w <= 4 && patch_h <= 4 ) {
		crosscorr_multi_slice<4,4><<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch,
			temp1.data, temp1.pitch, temp2.data, temp2.pitch,
			pitch_in.w, pitch_in.h, patch_mean, patch_std, 1.0f/float(patch_real_numel));
	}
	else if ( patch_w <= 8 && patch_h <= 8 ) {
		crosscorr_multi_slice<8,8><<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch,
			temp1.data, temp1.pitch, temp2.data, temp2.pitch,
			pitch_in.w, pitch_in.h, patch_mean, patch_std, 1.0f/float(patch_real_numel));
	}
	else if ( patch_w <= 8 && patch_h <= 16 ) {
		crosscorr_multi_slice<8,16><<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch,
			temp1.data, temp1.pitch, temp2.data, temp2.pitch,
			pitch_in.w, pitch_in.h, patch_mean, patch_std, 1.0f/float(patch_real_numel));
	}
	else if ( patch_w <= 16 && patch_h <= 8 ) {
		crosscorr_multi_slice<16,8><<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch,
			temp1.data, temp1.pitch, temp2.data, temp2.pitch,
			pitch_in.w, pitch_in.h, patch_mean, patch_std, 1.0f/float(patch_real_numel));
	}
	else if ( patch_w <= 16 && patch_h <= 16 ) {
		crosscorr_multi_slice<16,16><<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch,
			temp1.data, temp1.pitch, temp2.data, temp2.pitch,
			pitch_in.w, pitch_in.h, patch_mean, patch_std, 1.0f/float(patch_real_numel));
	}
	crosscorr_normit_slice<<< grid, block >>>(pitch_out.data, pitch_out.pitch, temp1.data, temp1.pitch, temp2.data, temp2.pitch,
		img_w, img_h, patch_mean, patch_std, 1.0f/float(patch_real_numel));
}


// Window sum and sum-sq function. Maximum window size of 16x16.
__global__ void window_sum_sum2_slice(float *din, int din_pitch, float *sum, int sum_pitch, float *sum2, int sum2_pitch, int img_w, int img_h, int patch_w, int patch_h)
{
	// Load din data into shared cache.
	__shared__ float data[ 1024 ];
	const int data_width = 32;	// Width of row in shared memory
	const int data_size = IMUL(data_width, blockDim.y);
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.

	// Read in data.
	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;		// Shared thread memory index.
	const float *my_din = PITCH_GET(din, din_pitch, y);				// Calculate my row of din.
	data[ix] = my_din[x];
	// Read in right-ward entry.
	const bool right_on = x + blockDim.x < img_w && threadIdx.x < patch_w;
	data[ix + blockDim.x] = right_on ? my_din[x + blockDim.x] : 0.0f;
	// Read in downward entries.
	if ( threadIdx.y < patch_h && y + blockDim.y < img_h ) {
		// Read in upper entries
		const float *in_up = PITCH_GET(din, din_pitch, y + blockDim.y);
		data[ ix + data_size ] = in_up[x];
		// Read in right-ward entry.
		data[ ix + data_size + blockDim.x ] = right_on ? in_up[x + blockDim.x] : 0.0f;
	} else {
		data[ ix + data_size ] = data[ ix + data_size + blockDim.x ] = 0.0f;
	}

	__syncthreads();

	// Compute my convolution result.
	float sumf = 0.0f, sumf2 = 0.0f;
#pragma unroll
	for ( int i=0; i<patch_h; ++i ) {
#pragma unroll
		for ( int j=0; j<patch_w; ++j ) {
			const float d = data[ ix + IMUL(data_width, i) + j ];
			sumf += d;
			sumf2 += d * d;
		}
	}

	// Write out results.
	float *sum_out = PITCH_GET(sum, sum_pitch, y);
	sum_out[x] = sumf;
	float *sum2_out = PITCH_GET(sum2, sum2_pitch, y);
	sum2_out[x] = sumf2;
}

// Compute windos sum and sum-squareds. Maximum window size is 16x16.
void svlCudaWindowSumSumSq(const svlCudaPitch *pin, svlCudaPitch *sum, svlCudaPitch *sum2, int win_w, int win_h, int img_w, int img_h)
{
	// TODO paulb: Make sum2 optional, in which case it uses a simpler function or something.
	if ( img_w == -1 ) img_w = pin->w;
	if ( img_h == -1 ) img_h = pin->h;
	dim3 block( 16, 16 );
	dim3 grid( iDivUp(img_w, block.x), iDivUp(img_h, block.y) );
	window_sum_sum2_slice<<< grid, block >>>(pin->data, pin->pitch, sum->data, sum->pitch, sum2->data, sum2->pitch, img_w, img_h, win_w, win_h);
}

// Do simple normalization using data plus corresponding sum and sum-squared images
void svlCudaPatchNormalize(svlCudaPitch *pout, svlCudaPitch *sum, svlCudaPitch *sum2, float patch_mean, float patch_std, int patch_N, int img_w, int img_h)
{
	if ( img_w == -1 ) img_w = pout->w;
	if ( img_h == -1 ) img_h = pout->h;
	dim3 block( 16, 16 );
	dim3 grid( iDivUp(img_w, block.x), iDivUp(img_h, block.y) );
	crosscorr_normit_slice<<< grid, block >>>(pout->data, pout->pitch, sum->data, sum->pitch, sum2->data, sum2->pitch,
		img_w, img_h, patch_mean, patch_std, 1.0f/float(patch_N)); 
}

// 2D convolution with normalization via pre-summed row-integral images.
template<int NUM_COLS>
__global__ void conv2d_halfT_normed_v2_slice(float *din, int din_pitch, float *res, int res_pitch,
											 float *integ, int integ_pitch, float *integ2, int integ2_pitch,
											 int data_w, int data_h, float patch_mean, float patch_inv_std, float N_inv, int NUM_ROWS)
{
	// Load din data into shared cache.
	__shared__ float data[ 1024 ];
	__shared__ float integd[ 512 ];
	__shared__ float integd2[ 512 ];

	const int data_width = 32;
	const int data_size = IMUL(data_width, blockDim.y);
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.

	// Read in data.
	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;		// Shared thread memory index.
	const float *my_din = PITCH_GET(din, din_pitch, y);				// Calculate my row of din.
	data[ix] = my_din[x];
	// Read in rightward entry.
	const bool right_on = x + blockDim.x < data_w;
	data[ix + blockDim.x] = right_on ? my_din[x + blockDim.x] : 0.0f;

	// Read in integral data (should be coalesced this way).
	const int i_ix = IMUL(16, threadIdx.y) + threadIdx.x;
	const float *integf = PITCH_GET(integ, integ_pitch, y);
	integd[i_ix] = integf[x];
	const float *integ2f = PITCH_GET(integ2, integ2_pitch, y);
	integd2[i_ix] = integ2f[x];

	// Read in upward data.
	if ( threadIdx.y < NUM_ROWS ) {
		if ( y + blockDim.y < data_h ) {
			// Read in upper entries
			const float *in_up = PITCH_GET(din, din_pitch, y + blockDim.y);
			data[ ix + data_size ] = in_up[x];
			// Read in rightward entry.
			data[ ix + data_size + blockDim.x ] = right_on ? in_up[x + blockDim.x] : 0.0f;

			// Integral data.
			const float *integf_up = PITCH_GET(integ, integ_pitch, y + blockDim.y);
			integd[ i_ix + 256 ] = integf_up[x];
			const float *integ2f_up = PITCH_GET(integ2, integ2_pitch, y + blockDim.y);
			integd2[ i_ix + 256 ] = integ2f_up[x];
		} else {
			data[ ix + data_size ] = data[ ix + data_size + blockDim.x ] = 0.0f;
			// Integral data.
			integd[ i_ix + 256 ] = integd2[ i_ix + 256 ] = 0.0f;
		}
	}

	__syncthreads();

	// Compute my convolution result.
	float gf = 0.0f, window_sum = 0.0f, window_sq_sum = 0.0f;
#pragma unroll
	for ( int i=0; i<NUM_ROWS; ++i ) {
		window_sum += integd[ i_ix + IMUL(16, i) ];
		window_sq_sum += integd2[ i_ix + IMUL(16, i) ];
#pragma unroll
		for ( int j=0; j<NUM_COLS; ++j ) {
			gf += kern_patch[ IMUL(NUM_COLS,i) + j ] * data[ ix + IMUL(data_width,i) + j ];
		}
	}

	// Calculate my row of the results.
	float *my_res = PITCH_GET(res, res_pitch, y);
	const float window_mean = window_sum * N_inv;
	const float window_sq_mean = window_sq_sum * N_inv;
	my_res[x] = __fdividef( gf - patch_mean * window_sum, sqrtf(window_sq_mean - window_mean*window_mean) ) * patch_inv_std * N_inv;
}

// Normalized cross correlation with half-T convolution.
void svlCudaNormCCorr2d_v3(const svlCudaPitch *pin, svlCudaPitch *pout, float *patch, float *mask, svlCudaPitch *integ, svlCudaPitch *integ2,
						   int patch_w, int patch_h, int patch_real_numel, float patch_mean, float patch_std, int img_w, int img_h, bool cached)
{
	// Copy patches over to GPU.
	if ( !cached ) {
		CUDA_SAFE_CALL( cudaMemcpyToSymbol( kern_patch, patch, patch_w*patch_h*sizeof(float), 0, cudaMemcpyHostToDevice) );
		//CUDA_SAFE_CALL( cudaMemcpyToSymbol( mask_patch, mask, patch_w*patch_h*sizeof(float), 0, cudaMemcpyHostToDevice) );
	} else {
		CUDA_SAFE_CALL( cudaMemcpyToSymbol( kern_patch, patch, patch_w*patch_h*sizeof(float), 0, cudaMemcpyDeviceToDevice) );
		//CUDA_SAFE_CALL( cudaMemcpyToSymbol( mask_patch, mask, patch_w*patch_h*sizeof(float), 0, cudaMemcpyDeviceToDevice) );
	}

	// Convolve.
	dim3 block(16, 16);
	if ( img_w == -1 ) img_w = pin->w;
	if ( img_h == -1 ) img_h = pin->h;
	dim3 grid( iDivUp(img_w, block.x), iDivUp(img_h, block.y) );

	patch_std = 1.0f / patch_std; // Pre-reciprocalize standard deviation

	float N_inv = 1.0f / float(patch_real_numel);
	if ( patch_w == 4 ) {
		conv2d_halfT_normed_v2_slice <4> <<< grid, block >>>(pin->data, pin->pitch, pout->data, pout->pitch,
			integ->data, integ->pitch, integ2->data, integ2->pitch, img_w, img_h, patch_mean, patch_std, N_inv, patch_h);
	} else if ( patch_w == 8 ) {
		conv2d_halfT_normed_v2_slice <8> <<< grid, block >>>(pin->data, pin->pitch, pout->data, pout->pitch,
			integ->data, integ->pitch, integ2->data, integ2->pitch, img_w, img_h, patch_mean, patch_std, N_inv, patch_h);
	} else if ( patch_w == 16 ) {
		conv2d_halfT_normed_v2_slice <16> <<< grid, block >>>(pin->data, pin->pitch, pout->data, pout->pitch,
			integ->data, integ->pitch, integ2->data, integ2->pitch, img_w, img_h, patch_mean, patch_std, N_inv, patch_h);
	} else {
		cout << "svlCudaNormCCorr2d_v3: Mis-padded patch.\n"; assert(false);
	}
}

