/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlAlgorithms.cpp
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
**
** DESCRIPTION:
**  Class implementing generic algorithms (some in template form) used across
**  the SVL codebase.
**
*****************************************************************************/

#include "svlAlgorithms.h"

void svlAlgorithms::computeNNNandRes(vector<svlPoint3d> &P, vector<svlPoint3d> &N, vector<vector<int> > &allNN, svlSortVectorDI *resid) {
	// Compute normals from nearest neighbors, and compute optional residuals.
	// P: vector of 3D points
	// N: corresponding vector of 3D normals, filled by the algorithm
	// allNN: double vector containing indices into P of neighbors of corresponding first-index elements
	// resid: pointer to IxVector for storing residuals for fast sorting. NULL to not calculate residuals.

	CvMat* PPT = cvCreateMat(3,3,CV_32FC1); // (P-C)*(P-C)^T
	CvMat* U = cvCreateMat(3,3,CV_32FC1); // left orthogonal matrix
	CvMat* S = cvCreateMat(3,3,CV_32FC1); // singular values
	CvMat* l = cvCreateMat(3,1,CV_32FC1); // vector

	int n_pts = (int)P.size();
	if ( (int)N.size() != n_pts ) {
		N.resize(n_pts); // I hope this allocates any needed extra space ...
	}
	if ( resid != NULL ) {
		resid->clear();
		resid->reserve(n_pts);
	}
	// compute first-pass nearest neighbors
	svlPoint3d cent, p, n;
	N = P; // copy
	for ( int i=0; i<n_pts; ++i ) {
		// compute centroid of vectors
		cent = 0.0;
		for ( vector<int>::const_iterator j=allNN[i].begin(); j!=allNN[i].end(); ++j ) {
			cent += P[*j];
		}
		cent /= (double)allNN[i].size();

		// collect covariance matrix
		cvZero(PPT);
		for ( vector<int>::const_iterator j=allNN[i].begin(); j!=allNN[i].end(); ++j ) {
			svlPoint3d pt = P[*j] - cent;
			CV_MAT_ELEM(*PPT,float,0,0) += (float)(pt.x*pt.x);
			CV_MAT_ELEM(*PPT,float,0,1) += (float)(pt.x*pt.y);
			CV_MAT_ELEM(*PPT,float,0,2) += (float)(pt.x*pt.z);
			CV_MAT_ELEM(*PPT,float,1,1) += (float)(pt.y*pt.y);
			CV_MAT_ELEM(*PPT,float,1,2) += (float)(pt.y*pt.z);
			CV_MAT_ELEM(*PPT,float,2,2) += (float)(pt.z*pt.z);
		}
		CV_MAT_ELEM(*PPT,float,1,0) = CV_MAT_ELEM(*PPT,float,0,1);
		CV_MAT_ELEM(*PPT,float,2,0) = CV_MAT_ELEM(*PPT,float,0,2);
		CV_MAT_ELEM(*PPT,float,2,1) = CV_MAT_ELEM(*PPT,float,1,2); // fill symmetric other half of matrix

		// caculate SVD and save vectors
		cvSVD(PPT,S,U,NULL,0);
		N[i].x = CV_MAT_ELEM(*U,float,0,2);
		N[i].y = CV_MAT_ELEM(*U,float,1,2);
		N[i].z = CV_MAT_ELEM(*U,float,2,2); // save new normal vector
		if ( resid != NULL ) {
			resid->push_back(fabs(N[i].dot(P[i] - cent)));
		}
	}

	//~
	cvReleaseMat(&PPT);
	cvReleaseMat(&U);
	cvReleaseMat(&S);
	cvReleaseMat(&l);
}

void svlAlgorithms::computeNNNandResMS(vector<svlPoint3d> &P, vector<svlPoint3d> &N, vector<vector<int> > &allNN, svlSortVectorDI *resid, int n_iter) {
	// Compute normals from nearest neighbors, and compute optional residuals.
	// P: vector of 3D points
	// N: corresponding vector of 3D normals, filled by the algorithm
	// allNN: double vector containing indices into P of neighbors of corresponding first-index elements
	// resid: pointer to IxVector for storing residuals for fast sorting. NULL to not calculate residuals.
//#define NN_DEBUG_OUT
	CvMat* PPT = cvCreateMat(3,3,CV_32FC1); // (P-C)*(P-C)^T
	CvMat* U = cvCreateMat(3,3,CV_32FC1); // left orthogonal matrix
	CvMat* S = cvCreateMat(3,3,CV_32FC1); // singular values
	CvMat* l = cvCreateMat(3,1,CV_32FC1); // vector

	int n_pts = (int)P.size();
	if ( (int)N.size() != n_pts ) {
		N.resize(n_pts); // I hope this allocates any needed extra space ...
	}
	if ( resid != NULL ) {
		resid->clear();
		resid->reserve(n_pts);
	}
	// compute first-pass nearest neighbors
	svlPoint3d cent, n, pt;
	N = P; // copy
	for ( int i=0; i<n_pts; ++i ) {
		// compute centroid of vectors
		cent = 0.0;
		for ( vector<int>::const_iterator j=allNN[i].begin(); j!=allNN[i].end(); ++j ) {
			cent += P[*j];
		}
		cent /= (double)allNN[i].size();

		// collect covariance matrix
		cvZero(PPT);
		for ( vector<int>::const_iterator j=allNN[i].begin(); j!=allNN[i].end(); ++j ) {
			pt = P[*j] - cent;
			CV_MAT_ELEM(*PPT,float,0,0) += (float)(pt.x*pt.x);
			CV_MAT_ELEM(*PPT,float,0,1) += (float)(pt.x*pt.y);
			CV_MAT_ELEM(*PPT,float,0,2) += (float)(pt.x*pt.z);
			CV_MAT_ELEM(*PPT,float,1,1) += (float)(pt.y*pt.y);
			CV_MAT_ELEM(*PPT,float,1,2) += (float)(pt.y*pt.z);
			CV_MAT_ELEM(*PPT,float,2,2) += (float)(pt.z*pt.z);
		}
		CV_MAT_ELEM(*PPT,float,1,0) = CV_MAT_ELEM(*PPT,float,0,1);
		CV_MAT_ELEM(*PPT,float,2,0) = CV_MAT_ELEM(*PPT,float,0,2);
		CV_MAT_ELEM(*PPT,float,2,1) = CV_MAT_ELEM(*PPT,float,1,2); // fill symmetric other half of matrix

		// caculate SVD and save vectors
		cvSVD(PPT,S,U,NULL,0);
		N[i].x = CV_MAT_ELEM(*U,float,0,2);
		N[i].y = CV_MAT_ELEM(*U,float,1,2);
		N[i].z = CV_MAT_ELEM(*U,float,2,2); // save new normal vector
		if ( resid != NULL ) {
			resid->push_back(fabs(N[i].dot(P[i] - cent)));
		}
	}

	if ( n_iter == 0 )
		return;
#ifdef NN_DEBUG_OUT
	FILE *fo = fopen("NN_-1.dat","wb");
	for ( int i=0; i<(int)N.size(); ++i ) {
		fwrite(&N[i].x,sizeof(double),1,fo);
		fwrite(&N[i].y,sizeof(double),1,fo);
		fwrite(&N[i].z,sizeof(double),1,fo);
	}
#endif

	float gf;
	vector<float> gains(allNN[0].size(),0.0);
	vector<svlPoint3d> N2(N); // dupe N
	for ( int iter=0; iter<n_iter; ++iter ) {
		if ( iter > 0 )
			N = N2;
		for ( int i=0; i<n_pts; ++i ) {
			// compute gains
			n = N[i];
			gf = 0.0;
			for ( int j=0; j<(int)allNN[i].size(); ++j ) {
				//gf += (gains[j] = 1.0 - (float)fabs(n.dot(N[allNN[i][j]]))); // need some math wrapping that function
				gf += (gains[j] = (float)exp( 5*( fabs(n.dot(N[allNN[i][j]])) - 1.0 ) ) );
			}

			// correct to gains, normalized to sum to 1
			//float factor = (float)gains.size() - 1.0;
			for ( int j=0; j<(int)gains.size(); ++j ) {
				//gains[j] = (gf-gains[j])/gf / factor;
				gains[j] /= gf;
			}

			// compute weighted centroid of vectors
			cent = 0.0;
			for ( int j=0; j<(int)allNN[i].size(); ++j ) {
				cent += gains[j] * P[allNN[i][j]];
			}

			// collect covariance matrix
			cvZero(PPT);
			for ( int j=0; j<(int)allNN[i].size(); ++j ) {
				pt = P[allNN[i][j]] - cent;
				CV_MAT_ELEM(*PPT,float,0,0) += gains[j] * (float)(pt.x*pt.x);
				CV_MAT_ELEM(*PPT,float,0,1) += gains[j] * (float)(pt.x*pt.y);
				CV_MAT_ELEM(*PPT,float,0,2) += gains[j] * (float)(pt.x*pt.z);
				CV_MAT_ELEM(*PPT,float,1,1) += gains[j] * (float)(pt.y*pt.y);
				CV_MAT_ELEM(*PPT,float,1,2) += gains[j] * (float)(pt.y*pt.z);
				CV_MAT_ELEM(*PPT,float,2,2) += gains[j] * (float)(pt.z*pt.z);
			}
			CV_MAT_ELEM(*PPT,float,1,0) = CV_MAT_ELEM(*PPT,float,0,1);
			CV_MAT_ELEM(*PPT,float,2,0) = CV_MAT_ELEM(*PPT,float,0,2);
			CV_MAT_ELEM(*PPT,float,2,1) = CV_MAT_ELEM(*PPT,float,1,2); // fill symmetric other half of matrix

			// caculate SVD and save vectors
			cvSVD(PPT,S,U,NULL,0);
			N2[i].x = CV_MAT_ELEM(*U,float,0,2);
			N2[i].y = CV_MAT_ELEM(*U,float,1,2);
			N2[i].z = CV_MAT_ELEM(*U,float,2,2); // save new normal vector
			if ( iter == n_iter-1 && resid != NULL )
				(*resid)[i].first = fabs(N2[i].dot(P[i]-cent));
		}
#ifdef NN_DEBUG_OUT
		char clift[32];
		sprintf(clift,"NN_%d.dat",iter);
		fo = fopen(clift,"wb");
		for ( int i=0; i<(int)N2.size(); ++i ) {
			fwrite(&N2[i].x,sizeof(double),1,fo);
			fwrite(&N2[i].y,sizeof(double),1,fo);
			fwrite(&N2[i].z,sizeof(double),1,fo);
		}
#endif
	}
	N = N2;

	//~
	cvReleaseMat(&PPT);
	cvReleaseMat(&U);
	cvReleaseMat(&S);
	cvReleaseMat(&l);
}

void svlAlgorithms::computeNormalMats(CvMat *_X, CvMat *_Y, CvMat *_Z, CvMat *_nX, CvMat *_nY, CvMat *_nZ, int n) {
#if 1
	int n2 = n>>1, ct;
	svlPoint3d t, p;
	CvMat *T = cvCreateMat(3,3,CV_32FC1),
		  *U = cvCreateMat(3,3,CV_32FC1),
		  *S = cvCreateMat(3,3,CV_32FC1);
	for ( int r=0; r<_X->rows; ++r ) {
		for ( int c=0; c<_X->cols; ++c ) {
			t = 0;
			cvZero(T);
			ct = 0;
			for ( int r1=std::max(0,r-n2); r1<std::min(_X->rows,r+n2+1); ++r1 ) {
				for ( int c1=std::max(0,c-n2); c1<std::min(_X->cols,c+n2+1); ++c1 ) {
					t += svlPoint3d(CV_MAT_ELEM(*_X,float,r1,c1),
						CV_MAT_ELEM(*_Y,float,r1,c1),
						CV_MAT_ELEM(*_Z,float,r1,c1));
					++ct;
				}
			}
			t /= (double)ct;
			for ( int r1=std::max(0,r-n2); r1<std::min(_X->rows,r+n2+1); ++r1 ) {
				for ( int c1=std::max(0,c-n2); c1<std::min(_X->cols,c+n2+1); ++c1 ) {
					p = svlPoint3d(CV_MAT_ELEM(*_X,float,r1,c1),
						CV_MAT_ELEM(*_Y,float,r1,c1),
						CV_MAT_ELEM(*_Z,float,r1,c1)) - t;
					CV_MAT_ELEM(*T,float,0,0) += (float)(p.x*p.x);
					CV_MAT_ELEM(*T,float,0,1) += (float)(p.x*p.y);
					CV_MAT_ELEM(*T,float,0,2) += (float)(p.x*p.z);
					CV_MAT_ELEM(*T,float,1,1) += (float)(p.y*p.y);
					CV_MAT_ELEM(*T,float,1,2) += (float)(p.y*p.z);
					CV_MAT_ELEM(*T,float,2,2) += (float)(p.z*p.z);
				}
			}
			CV_MAT_ELEM(*T,float,1,0) = CV_MAT_ELEM(*T,float,0,1);
			CV_MAT_ELEM(*T,float,2,0) = CV_MAT_ELEM(*T,float,0,2);
			CV_MAT_ELEM(*T,float,2,1) = CV_MAT_ELEM(*T,float,1,2);

			cvSVD(T,S,U,NULL,0);
			CV_MAT_ELEM(*_nX,float,r,c) = CV_MAT_ELEM(*U,float,0,2);
			CV_MAT_ELEM(*_nY,float,r,c) = CV_MAT_ELEM(*U,float,1,2);
			CV_MAT_ELEM(*_nZ,float,r,c) = CV_MAT_ELEM(*U,float,2,2);
		}
	}
#else
	list<vector<svlPoint3d> > Cols(n,vector<svlPoint3d>(n,svlPoint3d(0.0)));
	vector<svlPoint3d> ts(n,svlPoint3d(0.0));
	svlPoint3d t(0.0), p;
	CvMat *T = cvCreateMat(3,3,CV_32FC1),
		  *U = cvCreateMat(3,3,CV_32FC1),
		  *S = cvCreateMat(3,3,CV_32FC1);

	svlCircularIterator<list<vector<svlPoint3d> >,vector<svlPoint3d> > Cit(&Cols), Cit1(&Cols);
	svlCircularIterator<vector<svlPoint3d>,svlPoint3d> itt(&ts), itt1(&ts);
	int n2 = n>>1, ix, rows = _X->rows, cols = _X->cols, beat;
	cout << "n2: " << n2 << endl;
	for ( int r=0; r<rows; ++r ) {
		// init matrices and vectors
		ts = vector<svlPoint3d>(n,svlPoint3d(0.0));
		Cols = list<vector<svlPoint3d> >(n,vector<svlPoint3d>(n,svlPoint3d(0.0)));

		Cit = Cols.begin();
		itt = ts.begin();
		int r1 = std::max(0,r-n2), r2 = std::min(_X->rows,r+n2+1);
		int rdiff = r2-r1;
		// Head computation
		int c = 0;
		for ( ; c<=n2; ++c ) {
			//cout << "c: " << c << endl;
			ix = 0;
			for ( int r_in=r1; r_in<r2; ++r_in ) {
				*itt += (*Cit)[ix++] = svlPoint3d(CV_MAT_ELEM(*_X,float,r_in,c),
										CV_MAT_ELEM(*_Y,float,r_in,c),
										CV_MAT_ELEM(*_Z,float,r_in,c));
			}
			*itt /= (double)ix; // mean
			++Cit;
			++itt;
		}

		// Main body computation
		for ( ; c<cols; ++c ) {
			// calculate mean vector t
			t = 0.0;
			itt1 = itt-1;
			beat = std::min(n,c);
			for ( int b=0; b<beat; ++b ) {
				t += *itt1;
				--itt1;
			}
			t /= (double)beat;
			cvZero(T);
			// calculate T*T' of points with t
			Cit1 = Cit-1;
			for ( int b=0; b<beat; ++b ) {
				for ( ix=0; ix<rdiff; ++ix ) {
					p = (*Cit1)[ix]-t;
					CV_MAT_ELEM(*T,float,0,0) += (float)(p.x*p.x);
					CV_MAT_ELEM(*T,float,0,1) += (float)(p.x*p.y);
					CV_MAT_ELEM(*T,float,0,2) += (float)(p.x*p.z);
					CV_MAT_ELEM(*T,float,1,1) += (float)(p.y*p.y);
					CV_MAT_ELEM(*T,float,1,2) += (float)(p.y*p.z);
					CV_MAT_ELEM(*T,float,2,2) += (float)(p.z*p.z);
				}
				--Cit1;
			}
			CV_MAT_ELEM(*T,float,1,0) = CV_MAT_ELEM(*T,float,0,1);
			CV_MAT_ELEM(*T,float,2,0) = CV_MAT_ELEM(*T,float,0,2);
			CV_MAT_ELEM(*T,float,2,1) = CV_MAT_ELEM(*T,float,1,2);

			cvSVD(T,S,U,NULL,0);
			CV_MAT_ELEM(*_nX,float,r,c-n2-1) = CV_MAT_ELEM(*U,float,0,2);
			CV_MAT_ELEM(*_nY,float,r,c-n2-1) = CV_MAT_ELEM(*U,float,1,2);
			CV_MAT_ELEM(*_nZ,float,r,c-n2-1) = CV_MAT_ELEM(*U,float,2,2);

			// read in next column of data
			ix = 0;
			*itt = 0.0;
			for ( int r_in=r1; r_in<r2; ++r_in ) {
				*itt += (*Cit)[ix++] = svlPoint3d(CV_MAT_ELEM(*_X,float,r_in,c),
										CV_MAT_ELEM(*_Y,float,r_in,c),
										CV_MAT_ELEM(*_Z,float,r_in,c));
			}
			*itt /= (double)ix; // mean
			++Cit;
			++itt;
		}

		// Tail computation
		for ( ; c<=cols+n2; ++c ) {
			// calculate mean vector t
			t = 0.0;
			itt1 = itt-1;
			beat = n-(c-cols);
			for ( int b=0; b<beat; ++b ) {
				t += *itt1;
				--itt1;
			}
			t /= (double)beat;
			cvZero(T);
			Cit1 = Cit-1; // look back through points just cached
			for ( int b=0; b<beat; ++b ) {
				for ( ix=0; ix<rdiff; ++ix ) {
					p = (*Cit1)[ix]-t;
					CV_MAT_ELEM(*T,float,0,0) += (float)(p.x*p.x);
					CV_MAT_ELEM(*T,float,0,1) += (float)(p.x*p.y);
					CV_MAT_ELEM(*T,float,0,2) += (float)(p.x*p.z);
					CV_MAT_ELEM(*T,float,1,1) += (float)(p.y*p.y);
					CV_MAT_ELEM(*T,float,1,2) += (float)(p.y*p.z);
					CV_MAT_ELEM(*T,float,2,2) += (float)(p.z*p.z);
				}
				--Cit1;
			}
			CV_MAT_ELEM(*T,float,1,0) = CV_MAT_ELEM(*T,float,0,1);
			CV_MAT_ELEM(*T,float,2,0) = CV_MAT_ELEM(*T,float,0,2);
			CV_MAT_ELEM(*T,float,2,1) = CV_MAT_ELEM(*T,float,1,2);
			cvSVD(T,S,U,NULL,0);
			CV_MAT_ELEM(*_nX,float,r,c-n2-1) = CV_MAT_ELEM(*U,float,0,2);
			CV_MAT_ELEM(*_nY,float,r,c-n2-1) = CV_MAT_ELEM(*U,float,1,2);
			CV_MAT_ELEM(*_nZ,float,r,c-n2-1) = CV_MAT_ELEM(*U,float,2,2);
		}
	}
#endif
}

