/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlGeneralObjectDetector.cpp
** AUTHOR(S):   Ian Goodfellow <ia3n@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**
*****************************************************************************/

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include <limits>

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

//#include "svlBase.h"
//#include "svlVision.h"
#include "svlGeneralObjectDetector.h"

using namespace std;


static void markUsedVars(CvDTreeNode * n, vector<bool> & used)
{
  if (n->left)
    markUsedVars(n->left, used);

  if (n->right)
    markUsedVars(n->right, used);


  if (n->split)
    {
      used.at(n->split->var_idx) = true;
      assert(!n->split->next);
    }
}

static void markUsedVars(CvBoostTree * t,  vector<bool> & used)
{
  if (t->get_root())
    markUsedVars((CvDTreeNode *)t->get_root(), used);
}

static void mapTree(CvDTreeNode * n, const vector<unsigned> & map)
{
  if (n->left)
    mapTree(n->left, map);

  if (n->right)
    mapTree(n->right, map);


  if (n->split)
    {
      n->split->var_idx = map.at(n->split->var_idx);
      assert(!n->split->next);
    }
}

static void mapTree(CvBoostTree * t, const vector<unsigned> & map)
{
  if (t->get_root())
    mapTree((CvDTreeNode *) t->get_root(), map);
}



// svlGeneralObjectDetector ----------------------------------------------

svlGeneralObjectDetector::svlGeneralObjectDetector(const char *name, int windowWidth, int windowHeight) :
  svlSlidingWindowDetector(name, windowWidth, windowHeight), _extractor(NULL)
{
    // do nothing
}

svlGeneralObjectDetector::svlGeneralObjectDetector(const svlGeneralObjectDetector& c) :
  svlSlidingWindowDetector(c), _extractor(c._extractor->clone())
{
    // do nothing
}

svlGeneralObjectDetector::~svlGeneralObjectDetector()
{
  if (_extractor)
    delete _extractor;
}

// configuration and access functions
bool svlGeneralObjectDetector::readFeatures(const char *filename)
{
  if (_extractor)
    delete _extractor;

  _extractor = svlDetectorFeatureExtractor::getInstance(filename);


    if (_extractor) {

      if (_featureVector)
	{
	  cvReleaseMat(&_featureVector);
	}
      _featureVector = cvCreateMat(1, (int)_extractor->numFeatures(), CV_32FC1);

      assert(_featureVector);

 
      _windowWidth = _extractor->windowWidth();
      _windowHeight = _extractor->windowHeight();
      return true;
    }
    
    return false;
}
 
bool svlGeneralObjectDetector::writeFeatures(const char *filename)
{
    return _extractor->write(filename);
}



void svlGeneralObjectDetector::computeImageFeatureVectors(vector<vector<double> >& fv,
							     const vector<CvPoint>& locations,
							     const vector<IplImage *>& images,
								 bool sparse)
{
  assert(_extractor);
  _extractor->extract(locations, images, fv,  sparse);
}

void svlGeneralObjectDetector::computePatchFeatureVector(vector<double> &fv,
    const vector<IplImage *>& images)
{
	_extractor->windowResponse(images, fv);	
}

void svlGeneralObjectDetector::trim() 
{
  /***************
   * IAN, I copied some of your code and created a getUsedFeaturesList() function
   * in svlSlidingWindowDetector; it might make sense to decompose things out once you've
   * finished writing the code here.
   * I just wanted a quick-and-dirty trick to speed some things up, and it worked, thanks :)
   * I also copied the markUsedVars functions... I appologize for doing it in such a lame way,
   * but I didn't want to mess up anything you were doing, so I figured having two copies for
   * now is probably ok.
   *                      -- Olga
   *************/

  assert(_extractor);
  assert(_model);


  vector<bool> used(_extractor->numFeatures());

  const CvMat * active = NULL;

#if OPENCV_DEVEL
  active = _model->get_active_vars();
#else
  SVL_LOG(SVL_LOG_FATAL, "General object detector trimming is only supported using the SVN copy of OpenCV. Please use the latest copy of OpenCV and define OPENCV_DEVEL to 1 in your copy of svl's make.local file. Otherwise you are limited to using the patchdictionary feature extractor and trimming it with the trimDictionary.pl script")
#endif

  for (int i = 0; i < active->cols; i++)
    used[CV_MAT_ELEM(*active,int,0,i)] = true;


  cout << "Pruned extractor from "<<_extractor->numFeatures()<< "features to ";

  //trim extractor
  svlDetectorFeatureExtractor * trimmed = _extractor->getPrunedExtractor(used);
  delete _extractor;
  _extractor = trimmed;



  cout << _extractor->numFeatures() << " features." << endl;

}
