#include "svlVision.h"
//#include "svlDevel.h"
#include "svlISM.h"

//Constants for the default interest point detector
#define BLOCK_SIZE 3
#define APERTURE_SIZE 5
#define K 0.04
#define WIN_WIDTH 5
#define WIN_HEIGHT 5
#define THRESHOLD 0.001



double minn(double a, double b)
{
  if (a < b)
    return a;
  return b;
}

double maxx(double a, double b)
{
  if (a > b)
    return a;
  return b;
}


svlISM::svlISM(unsigned int width, unsigned int height, unsigned int numChannels, double ngcThreshold, svlInterestPointDetector * ipd) : _width(width), _height(height), _numChannels(numChannels), _phase(ACCUM), _ipd(ipd), _ngcThreshold(ngcThreshold), _showInterestPointsAccum(false)
{

  //If no interest point detector is supplied, construct a default one
  if (!_ipd)
    {
      _ownIpd = true;
      _ipd = new svlHarrisCornerDetector(BLOCK_SIZE, APERTURE_SIZE, K, WIN_WIDTH, WIN_HEIGHT, THRESHOLD);
    }
  else
    {
      _ownIpd = false;
    }

  for (unsigned i = 0; i < numChannels; i++)
    {
      _patches.push_back(vector<svlISMpatch *> () );
      _offsets.push_back(vector<vector<CvPoint> > () );
      _negatives.push_back(vector<unsigned int> () );
      _occurrances.push_back(vector<vector<svlOccurrance> > () );
    }

  if (width %2 == 0 || height % 2 == 0)
    SVL_LOG(SVL_LOG_FATAL, "svlISM must have odd patch dimensions");




  _halfWidth = (width-1) / 2;
  _halfHeight = (height-1) / 2;
}


svlISM::~svlISM()
{
  if (_ownIpd)
    delete _ipd;

  destroyPatches();
}

void svlISM::learnPatches(const vector<IplImage *> & channels)
{
  assert(channels.size() == _numChannels);
  assert(_phase == ACCUM);

  for (unsigned ch = 0; ch < _numChannels; ch++)
    {
      vector<CvPoint> interestPoints;

      _ipd->findInterestPoints(channels.at(ch), interestPoints);

      if (_showInterestPointsAccum)
	{
	  IplImage * color = colorImage(channels.at(ch));
	  CvScalar pink = cvScalar(255,0,255);
	  for (unsigned p = 0; p < interestPoints.size(); p++)
	    {

	      cvSet2D(color, interestPoints.at(p).y, interestPoints.at(p).x, pink);
	    }

#define IP_WIN "Interest Points"
	  cvNamedWindow(IP_WIN, CV_WINDOW_AUTOSIZE);
	  cvShowImage(IP_WIN, color);
	  cvReleaseImage(&color);
	  cvWaitKey(0);
	  cvDestroyWindow(IP_WIN);

	}

      vector<svlISMpatch *> newPatches;
      appendPatches(newPatches, channels.at(ch), interestPoints);


      
      unsigned orig = _patches.at(ch).size();


      //For each p in newPatches, add p to the codebook
      for (unsigned p = 0; p < newPatches.size(); p++)
	{
	  addPatch(newPatches.at(p), _patches.at(ch));
	}

      cout << "Processed "<<newPatches.size()<<" new patches" << endl;
      cout << "Dictionary grew by "<<((int) _patches.at(ch).size()- (int) orig)<<endl;
      cout << "Dictionary now contains "<<_patches.at(ch).size()<<" patches" << endl;
    }
}

void svlISM::addPatch(svlISMpatch * p, vector<svlISMpatch *> & dictionary)
{
  //Find the most similar existing patch
  double bestScore = -1;
  unsigned bestInd = 0;
  for (unsigned e = 0; e < dictionary.size(); e++)
    {
      double score = svlNGC(dictionary.at(e),p);
      if (score > bestScore)
	{
	  bestScore = score;
	  bestInd = e;
	}
    }

  if (bestScore < _ngcThreshold)
    dictionary.push_back(p);
  else
    {
      svlISMpatch * newPatch = svlMerge(p, dictionary.at(bestInd));
      delete dictionary.at(bestInd);
      dictionary.erase(dictionary.begin()+bestInd);
      delete p;

      addPatch(newPatch, dictionary);
    }
}

void svlISM::appendPatches(vector<svlISMpatch *> & output, const IplImage * img, const vector<CvPoint> & points) const
{
  for (unsigned i = 0; i < points.size(); i++)
    {
      svlISMpatch * temp = getPatch(img, points.at(i));

      if (temp)
	output.push_back(temp);
    }
}


svlISMpatch * svlISM::getPatch(const IplImage * img, const CvPoint & pt) const
{
  int minX = pt.x - _halfWidth;
  int minY = pt.y - _halfHeight;
  int maxX = pt.x + _halfWidth;
  int maxY = pt.y + _halfHeight;

  if (minX < 0)
    return NULL;

  if (minY < 0)
    return NULL;

  if (maxX >= img->width)
    return NULL;

  if (maxY >= img->height)
    return NULL;

  assert(_height == 7);

  CvMat * impatch = cvCreateMat(_height, _width, CV_32FC1);

  if (!impatch)
    return NULL;

  for (int x = minX; x <= maxX; x++)
    {
      for (int y = minY; y <= maxY; y++)
	{
	  cvmSet(impatch,y-minY,x-minX, cvGetReal2D(img, y,x) / 255.0);
	}
    }

  /*
  cvNamedWindow("Patch");
  cvShowImage("Patch",rval);
  cvWaitKey(0);
  cvDestroyWindow("Patch");
  */

  return new svlISMpatch(impatch);
}


/*
Trying to agglomerate with a straightforward interpretation of what the ISM
paper said to do
This ran out of memory. There is a space-time tradeoff but if I reduced the
memory consumption time would go up to O(n^3). Ouch.

void svlISM::agglomerate()
{
  _phase = COUNT;

  //We can do the agglomerative clustering separately for each channel
  for (unsigned ch = 0; ch < _numChannels; ch++)
    {
      SVL_LOG(SVL_LOG_MESSAGE, "Clustering "<<_patches.at(ch).size()<<" patches");

      //Each cluster is a list of patches that belong to that cluster
      vector<vector<CvMat *> > clusters(_patches.at(ch).size());

      //Initialize each cluster as containing a single patch
      for (unsigned p = 0; p < _patches.at(ch).size(); p++)
	{
	  clusters.at(p).push_back(_patches.at(ch).at(p));
	}

      //Clear out the old patches; they'll be repopulated at the end
      _patches.at(ch).clear();

      //The average Normalized Greyscale Correlation between a patch in
      //cluster i and cluster j is stored at (assuming wlog that j is the
      //greater index) at
      // avgNGC.at(i).at(j-i-1)
      // we do not store the correlation between i and j for i=j
      vector<vector<float> > avgNGC(clusters.size()-1);
      
      cout << "Allocating memory" << endl;

      for (unsigned i = 0; i < avgNGC.size(); i++)
	avgNGC.at(i).resize(clusters.size()-i-1);

      //Initialize the NGCs
      for (unsigned i = 0; i < avgNGC.size(); i++)
	{
	  cout << "Correlating with patch "<<i<<endl;
	  for (unsigned j = i + 1; j < clusters.size(); j++)
	    {
	      assert(clusters.at(i).size() == 1);
	      assert(clusters.at(j).size() == 1);

	      avgNGC.at(i).at(j-i-1)=  svlNGC(clusters.at(i).at(0), clusters.at(j).at(0));
	    }
	}



      //Repeatedly cluster until no merge will result in a cluster
      unsigned iter = 0;
      while (true)
	{
	  cout << "Iteration "<<++iter<<endl;
	  
	  unsigned int argmaxi = 0;
	  unsigned int argmaxj = 0;
	  double maxNGC = -1;

	  //Find the two most similar clusters
	  for (unsigned i = 0; i < avgNGC.size(); i++)
	    {
	      for (unsigned j = i + 1; j < clusters.size(); j++)
		{
		  double thisNGC = avgNGC.at(i).at(j-i-1);
		  if (thisNGC >= maxNGC)
		    {
		      argmaxi = i;
		      argmaxj = j;
		      maxNGC = thisNGC;
		    }
		}
	    }

	  //If no clusters are similar enough to merge, we're done clustering
	  if (maxNGC < _ngcThreshold)
	    break;

	  //We are now going to turn cluster argmaxi into the union
	  //of cluster argmaxi and argmaxj

	  double isize = clusters.at(argmaxi).size();
	  double jsize = clusters.at(argmaxj).size();
	  double iprop = isize / (isize+jsize);
	  double jprop = jsize / (isize+jsize);

	  //Update the avgNGC assuming that argmaxi now also contains the patches in argmaxj

	  //Go through all the rows that contain a single edge involving argmaxi
	  for (unsigned i = 0; i < argmaxi; i++)
	    {
	      //Update the avgNGC between a patch in i and a patch in argmaxi
	      avgNGC.at(i).at(argmaxi-i-1) = iprop * avgNGC.at(i).at(argmaxi-i-1) + jprop * avgNGC.at(i).at(argmaxj-i-1);
	    }

	  double j_argmaxj = 0;
	  
	  //Go through the row for argmaxi itself
	  for (unsigned j = argmaxi + 1; j < clusters.size(); j++)
	    {
	      if (j == argmaxj)
		continue;

	      if (j < argmaxj)
		j_argmaxj = avgNGC.at(j).at(argmaxj-j-1);
	      else
		j_argmaxj = avgNGC.at(argmaxj).at(j-argmaxj-1);

	      avgNGC.at(argmaxi).at(j-argmaxi-1) = iprop * avgNGC.at(argmaxi).at(j-argmaxi-1) + jprop * j_argmaxj;
	    }

	  //Copy the patches of argmaxj over to argmaxi
	  for (unsigned i = 0; i < clusters.at(argmaxj).size(); i++)
	    clusters.at(argmaxi).push_back(clusters.at(argmaxj).at(i));

	  //Delete all references to argmaxj
	  for (unsigned i = 0; i < argmaxj; i++)
	    avgNGC.at(i).erase(avgNGC.at(i).begin()+argmaxj-i-1);

	  if (argmaxj != clusters.size() -1)
	    avgNGC.erase(avgNGC.begin()+argmaxj);

	  clusters.erase(clusters.begin()+argmaxj);

	  
	}

      //Compute the center of each cluster and store these in the patch dictionary
      for (unsigned i = 0; i < clusters.size(); i++)
	{
	  if (clusters.at(i).size() == 1)
	    {
	      _patches.at(ch).push_back(clusters.at(i).at(0));
	    }
	  else
	    {
	      CvMat * temp = cvCreateMat(_height,_width, CV_32FC1);
	      cvZero(temp);

	      for (unsigned j = 0; j < clusters.at(i).size(); j++)
		{
		  cvmAdd(temp,clusters.at(i).at(j),temp);
		  cvReleaseMat(&clusters.at(i).at(j));
		}   

	      cvScale(temp,temp,1.0/double(clusters.at(i).size()));
	      _patches.at(ch).push_back(temp);
	    }
	}

      SVL_LOG(SVL_LOG_MESSAGE, "Clustering complete. "<<_patches.at(ch).size()<<" patches found");
    }
}

*/


void svlISM::learnOffsets(const vector<IplImage *> & channels, const svlObject2dFrame & labels)
{
  assert(_phase = COUNT);

  unsigned frame = _objects.size();

  _objects.push_back(labels);

  for (unsigned ch = 0; ch < _numChannels; ch++)
    {

      //cout << "A" << endl;
      vector<CvPoint> locations;
      vector<svlISMpatch *> impatches;
      getPatchesAndLocations(channels.at(ch), impatches, locations);


      for (unsigned i = 0; i < locations.size(); i++)
	{

	  //cout << "B" << endl;

	  //List of offset to all objects containing locations.at(i)
	  vector<CvPoint> curOffsets;
	  
	  double x = locations.at(i).x;
	  double y = locations.at(i).y;


	  for (unsigned j = 0; j < labels.size(); j++)
	    {
	      if (labels.at(j).hit(x,y))
		{
		  curOffsets.push_back(cvPoint(labels.at(j).x+labels.at(j).w/2.0-x, labels.at(j).y+labels.at(j).h/2.0-y));
		}
	    }
	  






	  for (unsigned j = 0; j < _patches.at(ch).size(); j++)
	    {
	      

		  
		  if (svlNGC(impatches.at(i), _patches.at(ch).at(j)) > _ngcThreshold)
		    {

		      _occurrances.at(ch).at(j).push_back(svlOccurrance(locations.at(i), frame));

		      if (curOffsets.size())
			{
			  //cout << "DAAA"<< endl;
			  for (unsigned k = 0; k < curOffsets.size(); k++)
			    _offsets.at(ch).at(j).push_back(curOffsets.at(k));
			  
		      
			}
		      else
			_negatives.at(ch).at(j)++;
		      
		    }
		  
		  
		  
	    }
	  
	  
	}
    }
}

#define CONDITIONAL_VOTE 0
svlObject2dFrame svlISM::getDetections(const vector<IplImage *> & channels)
{
  _votes.clear();
  _voteWeights.clear();
  _voteSources.clear();

#if CONDITIONAL_VOTE
  int w = channels.at(0)->width;
  int h = channels.at(0)->height;


#define DOWNSCALE_VOTE 10

  CvMat * posVotes = cvCreateMat(h/DOWNSCALE_VOTE,w/DOWNSCALE_VOTE,CV_32FC1);
  CvMat * negVotes = cvCreateMat(h/DOWNSCALE_VOTE,w/DOWNSCALE_VOTE,CV_32FC1);

  for (int i = 0; i < h / DOWNSCALE_VOTE; i++ )
    {
      for (int j = 0; j < w / DOWNSCALE_VOTE; j++)
	{
	  cvmSet(posVotes,i,j,1);
	  cvmSet(negVotes,i,j,1);
	}
    }
#endif
  

  for (unsigned ch = 0; ch < _numChannels; ch++)
    {
      
      vector<CvPoint> locations;
      vector<svlISMpatch *> impatches;
      getPatchesAndLocations(channels.at(ch), impatches, locations);
      

      //for (int x = _halfWidth; x + _halfWidth < channels.at(0)->width; x+=2)
      //{
      //for (int y = _halfHeight; y + _halfHeight < channels.at(0)->height; y+=2)
      //locations.push_back(cvPoint(x,y));
      //}
      
      //appendPatches(impatches, channels.at(0), locations);


      for (unsigned i = 0; i < locations.size(); i++)
	{
	  //cout << "Location: "<<i<<" of "<<locations.size()<<endl;

	  vector<unsigned> responders;
	  for (unsigned j = 0; j < _patches.at(ch).size(); j++)
	    {
	      if (svlNGC(impatches.at(i), _patches.at(ch).at(j)) > _ngcThreshold)
		{
		  responders.push_back(j);
		}
	    }

	  double patchProb = 1.0 / double(responders.size());

	  for (unsigned j = 0; j < responders.size(); j++)
	    {
	      unsigned p = responders.at(j);

	      //cout << "Responder: " << j << " of " << responders.size()<< endl;


#if CONDITIONAL_VOTE
	      {
		//Inside these braces is all silly Ia3n stuff

		double occurranceProb = 1.0 / double(_occurrances.at(ch).at(p).size());

		double weight = patchProb * occurranceProb;

		for (unsigned occ = 0; occ < _occurrances.at(ch).at(p).size(); occ++)
		  {
		    for (int x_orig = 0; x_orig < w; x_orig+= DOWNSCALE_VOTE)
		      {
			for (int y_orig = 0; y_orig < h; y_orig+= DOWNSCALE_VOTE)
			  {
			    #define INFO_DIST 30.0

			    double x_diff = _occurrances.at(ch).at(p).at(occ).x - x_orig;
			    double y_diff = _occurrances.at(ch).at(p).at(occ).y - y_orig;

			    if (x_diff*x_diff + y_diff * y_diff > INFO_DIST * INFO_DIST)
			      continue;
						  


			    int y_new = locations.at(i).y + y_orig - _occurrances.at(ch).at(p).at(occ).y;
			    
			    if (y_new >= h)
			      continue;
			    
			    if (y_new < 0)
			      continue;
			    
			    int x_new = locations.at(i).x + x_orig - _occurrances.at(ch).at(p).at(occ).x;

			    if (x_new >= w)
			      continue;
			    
			    if (x_new < 0)
			      continue;
			    
			    bool positive = false;

			    for (unsigned obj = 0; obj < _objects.at(_occurrances.at(ch).at(p).at(occ).frame).size();obj++)
			      {
				if (_objects.at(_occurrances.at(ch).at(p).at(occ).frame).at(obj).hit(x_orig,y_orig))
				  {
				    positive = true;
				    break;
				}
			      }

			    x_new /= DOWNSCALE_VOTE;
			    y_new /= DOWNSCALE_VOTE;

			    if (x_new >= posVotes->cols)
			      continue;

			    if (y_new >= negVotes->rows)
			      continue;

			    if (positive)
			      cvmSet(posVotes,y_new,x_new,cvmGet(posVotes,y_new,x_new) + weight);
			    else
			      cvmSet(negVotes,y_new,x_new,cvmGet(negVotes,y_new,x_new) + weight);
			  

			    //todo-- this needs cleanup. if we are doing the downscaled voting, should map a new pixel to an old pixel rather than vice versa. could also do antialiased rectangle rendering rather than pixel by pixel hit tests.
			    
			  }
		      }
		  }


	      }
#endif


	      double numNeg = _negatives.at(ch).at(p);
	      double numPos = _offsets.at(ch).at(p).size();

	      // P(location | object, patch) P(object | patch) P(patch | pixels)
	      double weight =  patchProb * (numPos / (numNeg+numPos))  / numPos;

	      for (unsigned k = 0; k < _offsets.at(ch).at(p).size(); k++)
		{

		  _votes.push_back(cvPoint(_offsets.at(ch).at(p).at(k).x+locations.at(i).x, _offsets.at(ch).at(p).at(k).y+locations.at(i).y));
		  _voteWeights.push_back(weight);
		  _voteSources.push_back(locations.at(i));

		}
	    }
	}      

      for (unsigned i = 0; i < impatches.size(); i++)
	delete impatches[i];

    }

  CvScalar yellow = cvScalar(0,255,255);
  
  IplImage * color = colorImage(channels.at(0));
  
  CvMat * voteImg = cvCreateMat(channels.at(0)->height, channels.at(0)->width, CV_32FC1);
  cvZero(voteImg);
  
  for (unsigned j = 0; j < _votes.size(); j++)
    {
      if ( 
	      (_votes.at(j).x >= 0) && 
	      (_votes.at(j).x <= color->width -1 )  && 
	       (_votes.at(j).y >= 0)  && 
	      (_votes.at(j).y < color->height))
	    {
	      //cout << _voteWeights.at(j) << endl;
	      cvmSet(voteImg,_votes.at(j).y,_votes.at(j).x, _voteWeights.at(j)+cvmGet(voteImg,_votes.at(j).y, _votes.at(j).x));
	    }
	}
      
      cout << "Minimum vote value: " << svlMin(voteImg) << endl;
      cout << "Maximum vote value: " << svlMax(voteImg) << endl;
      

      
      #define MAXFINDER_HALFWIDTH 10
      #define MAXFINDER_HALFHEIGHT 10

      /*
      double maxScore = 0;
      int maxI = 0;
      int maxJ = 0;

      for (int i = MAXFINDER_HALFHEIGHT; i < voteImg->rows-MAXFINDER_HALFHEIGHT; i++)
	{
	  for (int j = MAXFINDER_HALFWIDTH; j < voteImg->cols-MAXFINDER_HALFWIDTH; j++)
	    {
	      double vote = 0;

	      for (int ii = -MAXFINDER_HALFHEIGHT; ii <= MAXFINDER_HALFHEIGHT; ii++)
		{
		  for (int jj = -MAXFINDER_HALFWIDTH; jj <= MAXFINDER_HALFWIDTH; jj++)
		    {
		      vote += cvmGet(voteImg, i+ii,j+jj);
		    }
		}

	      if (vote >= maxScore)
		{
		  maxScore = vote;
		  maxI = i;
		  maxJ = j;
		}
	    }
	}

      cout << "Winning vote: " << maxScore << endl;
      */


      vector<CvPoint> maxima;
      
      svlFindLocalMaxima(voteImg, maxima, MAXFINDER_HALFWIDTH*2+1,MAXFINDER_HALFHEIGHT*2+1);


      vector<double> scores;

      for (int m = 0; m < maxima.size(); m++)
	{
	  double score = 0;
	  for (int i = - MAXFINDER_HALFHEIGHT; i <= MAXFINDER_HALFHEIGHT; i++)
	    for (int j = -MAXFINDER_HALFWIDTH; j <= MAXFINDER_HALFWIDTH; j++)
	      {
		if (maxima.at(m).y + i >= 0 && maxima.at(m).x+j >= 0 && maxima.at(m).y+i < voteImg->rows && maxima.at(m).x+j < voteImg->cols)
		score += cvmGet(voteImg, maxima.at(m).y+i,maxima.at(m).x+j);
	      }

	  if (score < 0.01)
	    {
	      maxima.erase(maxima.begin()+m);
	      m--;
	    }
	  else
	    scores.push_back(score);
	}

      double low = svlMin(voteImg);
      double high = svlMax(voteImg);

      for (int x = 0; x < color->width; x++)
	{
	  for (int y = 0; y < color->height; y++)
	    {
	      CvScalar pix = cvGet2D(color,y,x);
	      double alpha = (cvmGet(voteImg,y,x) - low) * (high-low);
	      for (unsigned j = 0; j < 3; j++)
		pix.val[j] =  alpha * yellow.val[j] + (1.0-alpha) * pix.val[j];
	      cvSet2D(color,y,x,pix);
	    }
	}

      svlObject2dFrame rval;


      for (unsigned m = 0; m < maxima.size(); m++)
	{

	  cout << "Vote for maximum "<<m<<"/"<<maxima.size()<<": "<<scores.at(m)<<endl;

	  
	  int centroid_x = maxima.at(m).x;
	  int centroid_y = maxima.at(m).y;
	  
	  IplImage * color_this = cvCloneImage(color);	  
	  
	  CvScalar red = cvScalar(0,0,255);
	  
	  
	  

	  for (int i = - 10; i <= 10; i++)
	    {
	      int x1 = centroid_x + i;
	      int y1 = centroid_y;
	      
	      if (x1 > 0 && y1 > 0 && x1 < color->width && y1 < color->height)
		cvSet2D(color_this,y1,x1,red);
	      
	      x1 = centroid_x;
	      y1 = centroid_y + i;
	      
	      if (x1 > 0 && y1 > 0 && x1 < color->width && y1 < color->height)
		cvSet2D(color_this,y1,x1,red);
	    }

	  

	  cvNamedWindow("Votes");
	  cvShowImage("Votes",color_this);

	  cvReleaseImage(&color_this);




	  IplImage * seg = colorImage(channels.at(0));
	  
	  unsigned minX = channels.at(0)->width-1;
	  unsigned maxX = 0;
	  unsigned minY = channels.at(0)->height-1;
	  unsigned maxY = 0;
	  


	  assert(_voteSources.size() == _votes.size());

	  for (unsigned i = 0; i < _votes.size(); i++)
	    {
	      int x = _votes.at(i).x;
	      int y = _votes.at(i).y;
	      
	      
	      if (x >= centroid_x - MAXFINDER_HALFWIDTH && x <= centroid_x + MAXFINDER_HALFWIDTH && y >= centroid_y - MAXFINDER_HALFHEIGHT && y <= centroid_y + MAXFINDER_HALFHEIGHT)
		{
		  unsigned xsource = _voteSources.at(i).x;
		  unsigned ysource = _voteSources.at(i).y;
		  
		  cvSet2D(seg, ysource, xsource, yellow);
		  
		  
		  if (xsource < minX)
		    minX = xsource;
		  
		  if (ysource < minY)
		    minY = ysource;
		  
		  if (xsource > maxX)
		    maxX = xsource;
		  
		  if (ysource > maxY)
		    maxY = ysource;
		}
	    }

	  
	  cvNamedWindow("Segmentation");
	  cvShowImage("Segmentation", seg);


	  //Resample for the "refined hypothesis"
	  
	  vector<CvPoint> votes2;
	  _voteWeights.clear();
	  vector<CvPoint> _voteSources2;
	  
	  for (unsigned ch = 0; ch < _numChannels; ch++)
	    {
	      vector<CvPoint> locations;
	      vector<svlISMpatch *> impatches;
	      
	      for (unsigned x = minX; x <= maxX; x++)
		{
		  for (unsigned y = minY; y <= maxY; y++)
		    {
		      locations.push_back(cvPoint(x,y));
		      impatches.push_back(getPatch(channels.at(ch),cvPoint(x,y)));
		    }
		}
	      



	      for (unsigned i = 0; i < locations.size(); i++)
		{
		  //cout << "Location: "<<i<<" of "<<locations.size()<<endl;
		  
		  vector<unsigned> responders;
		  for (unsigned j = 0; j < _patches.at(ch).size(); j++)
		    {
		      if (svlNGC(impatches.at(i), _patches.at(ch).at(j)) > _ngcThreshold)
			{
			  responders.push_back(j);
			}
		    }
		  
		  double patchProb = 1.0 / double(responders.size());
		  
		  for (unsigned j = 0; j < responders.size(); j++)
		    {
		      unsigned p = responders.at(j);
		      
		      double numNeg = _negatives.at(ch).at(p);
		      double numPos = _offsets.at(ch).at(p).size();
		      
		      // P(location | object, patch) P(object | patch) P(patch | pixels)
		      double weight =  patchProb * (numPos / (numNeg+numPos))  / numPos;

		      //cout << "I i=" <<i << " j="<<j<<" m="<<m<< endl;
		      assert( p < _offsets.at(ch).size());
		      assert( i < locations.size() );



		      bool debug = (i == 15965) && (j == 0) && (m == 37);

		      for (unsigned k = 0; k < _offsets.at(ch).at(p).size(); k++)
			{
			  if (debug)
			    cout << "I1" << endl;
			  CvPoint pt(cvPoint(_offsets.at(ch).at(p).at(k).x+locations.at(i).x, _offsets.at(ch).at(p).at(k).y+locations.at(i).y));
			  if (debug)
			    {
			      cout << "I2" << endl;
			      CvPoint * test = new CvPoint;
			      cout << "I2A" << endl;
			      delete test;
			      cout << "I2B" << endl;
			      test = new CvPoint[1000];
			      cout << "I2C" << endl;
			      delete [] test;
			      cout << "I2D" << endl;
			      cout << votes2.size() << endl;
			    }
			  votes2.push_back(pt);
			  if (debug)
			    cout << "I3" << endl;
			  _voteWeights.push_back(weight);
			  if (debug)
			    cout << "I4" << endl;
			  _voteSources2.push_back(locations.at(i));
			  if (debug)
			    cout << "I5" << endl;
			}
		      //cout << "J" << endl;
		    }
		}

	      for (unsigned p = 0; p < impatches.size(); p++)
		delete impatches[p];
	    }
	  
	  


	  cvZero(voteImg);

	  
	  for (unsigned j = 0; j < votes2.size(); j++)
	    {
	      if ( 
		  (votes2.at(j).x >= 0) && 
		  (votes2.at(j).x <= color->width -1 )  && 
		  (votes2.at(j).y >= 0)  && 
		  (votes2.at(j).y < color->height))
		{
		  //cout << _voteWeights.at(j) << endl;
		  cvmSet(voteImg,votes2.at(j).y,votes2.at(j).x, _voteWeights.at(j)+cvmGet(voteImg,votes2.at(j).y, votes2.at(j).x));
		}
	    }
	  
	  cout << "Minimum refined vote value: " << svlMin(voteImg) << endl;
	  cout << "Maximum refined vote value: " << svlMax(voteImg) << endl;
	  
	  
	  double maxScore = 0;
	  int maxI = 0;
	  int maxJ = 0;
	  
	  for (int i = MAXFINDER_HALFHEIGHT; i < voteImg->rows-MAXFINDER_HALFHEIGHT; i++)
	    {
	      for (int j = MAXFINDER_HALFWIDTH; j < voteImg->cols-MAXFINDER_HALFWIDTH; j++)
		{
		  double vote = 0;
		  
		  for (int ii = -MAXFINDER_HALFHEIGHT; ii <= MAXFINDER_HALFHEIGHT; ii++)
		    {
		      for (int jj = -MAXFINDER_HALFWIDTH; jj <= MAXFINDER_HALFWIDTH; jj++)
			{
			  vote += cvmGet(voteImg, i+ii,j+jj);
			}
		    }
		  
		  if (vote >= maxScore)
		    {
		      maxScore = vote;
		      maxI = i;
		      maxJ = j;
		    }
		}
	    }
	  

	  cout << "S" << endl;


	  cout << "Winning vote, refined : " << maxScore << endl;
	  
	  IplImage * color2 = cvCloneImage(color);
	  
	  cvZero(color2);
	  
	  scaleToRange(voteImg, 0, 0.75);
	  
	  for (int x = 0; x < color2->width; x++)
	    {
	      for (int y = 0; y < color2->height; y++)
		{
		  CvScalar pix = cvGet2D(color2,y,x);
		  double alpha = cvmGet(voteImg,y,x);
		  for (unsigned j = 0; j < 3; j++)
		    pix.val[j] =  alpha * yellow.val[j] + (1.0-alpha) * pix.val[j];
		  cvSet2D(color2,y,x,pix);
		}
	    }



	  
	  for (int i = - 10; i <= 10; i++)
	    {
	      int x1 = maxJ + i;
	      int y1 = maxI;
	      
	      if (x1 > 0 && y1 > 0 && x1 < color->width && y1 < color->height)
		cvSet2D(color2,y1,x1,red);
	      
	      x1 = maxJ;
	      y1 = maxI + i;
	      
	      if (x1 > 0 && y1 > 0 && x1 < color->width && y1 < color->height)
		cvSet2D(color2,y1,x1,red);
	    }
	  
	  cvNamedWindow("Refined votes");
	  cvShowImage("Refined votes",color2);
	  cvReleaseImage(&color2);
	  
	  

	  
	  
	  
	  IplImage * seg2 = colorImage(channels.at(0));
	  
	  
	  minX = channels.at(0)->width-1;
	  maxX = 0;
	  minY = channels.at(0)->height-1;
	  maxY = 0;
	  
	  assert(votes2.size() == _voteSources2.size());

	  for (unsigned i = 0; i < votes2.size(); i++)
	    {
	      int x = votes2.at(i).x;
	      int y = votes2.at(i).y;
	      
	      
	      
	      if (x >= maxJ - MAXFINDER_HALFWIDTH && x <= maxJ + MAXFINDER_HALFWIDTH && y >= maxI - MAXFINDER_HALFHEIGHT && y <= maxI + MAXFINDER_HALFHEIGHT)
		{
		 unsigned xsource = _voteSources2.at(i).x;
		 unsigned ysource = _voteSources2.at(i).y;
		  
		  cvSet2D(seg2, ysource, xsource, yellow);
		  
		  if (xsource < minX)
		    minX = xsource;
		  
		  if (ysource < minY)
		    minY = ysource;
		  
		  if (xsource > maxX)
		    maxX = xsource;
		  
		  if (ysource > maxY)
		    maxY = ysource;
		}
	    }
	  
	  cvNamedWindow("Refined Segmentation");
	  cvShowImage("Refined Segmentation", seg2);
	  
	  cvReleaseImage(&seg2);

#if CONDITIONAL_VOTE
	  assert(false);
	  CvMat * condProb = cvCreateMat(h/DOWNSCALE_VOTE,w/DOWNSCALE_VOTE,CV_32FC1);
	  IplImage * condProbImg = cvCreateImage(cvSize(w/DOWNSCALE_VOTE,h/DOWNSCALE_VOTE),IPL_DEPTH_8U,3);
	  
	  for (int i = 0; i < h / DOWNSCALE_VOTE; i++)
	    for (int j = 0; j < w / DOWNSCALE_VOTE; j++)
	      {
		double p = cvmGet(posVotes,i,j);
		double n = cvmGet(negVotes,i,j);
		double prob = p/(p+n);
		cvmSet(condProb,i,j, prob);
		
		CvScalar s = cvScalar(0,0,0);
		
		if (prob >= 0.5)
		  s.val[1] = (prob - 0.5) * 2.0 * 255.0;
		else
		  s.val[2] = (1.0-prob *2.0) * 255.0;
		
		cvSet2D(condProbImg,i,j,s);
	      }
	  
	  
	  resizeInPlace(&condProbImg, h, w, CV_INTER_NN);
	  
	  
	  
	  cvNamedWindow("Conditional Prob");
	  cvShowImage("Conditional Prob",condProbImg);
	  
	  
	  cvReleaseImage(&condProbImg);
	  cvReleaseMat(&posVotes);
	  cvReleaseMat(&negVotes);
	  cvReleaseMat(&condProb);
#endif
	  
	  cout << "Waiting" << endl;
	  //cvWaitKey(0);
	  cout << "Running" << endl;
	  

	  cvReleaseImage(&seg);
	  

	  svlObject2d obj(minX,minY,maxX-minX,maxY-minY,maxScore);

	  if (obj.w >= 500)
	    {
	      cout << "maxX: " << maxX << endl;
	      cout << "minX: " << minX << endl;
	    }

	  if (obj.h >= 300)
	    {
	      cout << "maxY: " << maxY << endl;
	      cout << "minY: " << minY << endl;
	    }


	  assert(obj.w < 500);
	  assert(obj.h < 300);

	  rval.push_back(obj);
	}

	  cvReleaseMat(&voteImg);
      cvReleaseImage(&color);


      return rval;
}

void svlISM::getPatchesAndLocations(const IplImage * img, vector<svlISMpatch *> & impatches, vector<CvPoint> & locations) const
{
  cout << "gpal A" << endl;
  impatches.clear();
  locations.clear();
  cout << "gpal B" << endl;



  _ipd->findInterestPoints(img,locations);

  cout << "gpal C" << endl;

  vector<CvPoint>::iterator i = locations.begin();

  cout << locations.size() << " locations" << endl;

  while (i != locations.end())
    {
      svlISMpatch * temp = getPatch(img, *i);

      if (temp)
	{
	  impatches.push_back(temp);
	  ++i;
	}
      else
	locations.erase(i);
    }
}


void svlISM::finishLearningPatches()
{
  _phase = COUNT;

  for (unsigned ch = 0; ch < _numChannels; ch++)
    {
      _offsets.at(ch).resize(_patches.at(ch).size());
      _negatives.at(ch).resize(_patches.at(ch).size());
      _occurrances.at(ch).resize(_patches.at(ch).size());

      for (unsigned n = 0; n < _negatives.at(ch).size(); n++)
	_negatives.at(ch).at(n) = 0;
    }
}


bool svlISM::write(const char *filename) const
{
  ofstream ofs(filename);

  ofs << "<ISM version=\"1\"\n"
      << " width=\"" << _width << "\"\n"
      << " height=\"" << _height << "\"\n"
      << " halfWidth=\"" << _halfWidth << "\"\n"
      << " halfHeight=\"" << _halfHeight << "\"\n"
      << " numChannels=\"" << _numChannels << "\"\n"
      << " phase=\"" << (int) _phase << "\"\n";

  ofs << " ownIpd=\"" << _ownIpd << "\"\n";
  ofs << " ngcThreshold=\"" << _ngcThreshold << "\"\n";
  ofs << " showInterestPointsAccum=\"" << _showInterestPointsAccum << "\"\n";
  ofs << " numFrames=\"" << _objects.size() << "\"\n";
  ofs << ">\n";

  _ipd->write(ofs);


  for (unsigned i = 0; i < _objects.size(); i++)
    writeObject2dFrame(ofs, _objects.at(i));


  for (unsigned ch = 0; ch < _numChannels; ch++)
    {

      ofs << "<Channel \n";
      ofs << " numEntries=\"" << _patches.at(ch).size() << "\"\n";
      ofs << " >\n";

      for (unsigned i = 0; i < _patches.at(ch).size(); i++)
	{
	  ofs << "<Patch \n";

	  if (_phase == COUNT)
	    {
	      ofs << " negatives=\"" << _negatives.at(ch).at(i) << "\"\n";
	    }

	  ofs << "> \n";

	  ofs << "<Pixels> \n";

	  _patches.at(ch).at(i)->writePixels(ofs);

	  ofs << "\n</Pixels> \n";



	  if (_phase == COUNT)
	    {
	      ofs << "<Offsets> \n";

	      for (unsigned j = 0; j < _offsets.at(ch).at(i).size(); j++)
		ofs << _offsets.at(ch).at(i).at(j).x << "\t" << _offsets.at(ch).at(i).at(j).y<<"\n";

	      ofs << "</Offsets> \n";

	      ofs << "<Occurrances> \n";

	      for (unsigned j = 0; j < _occurrances.at(ch).at(i).size(); j++)
		ofs << _occurrances.at(ch).at(i).at(j).x << "\t" << _occurrances.at(ch).at(i).at(j).y << "\t" << _occurrances.at(ch).at(i).at(j).frame << "\n";

	      ofs << "</Occurrances> \n";


	    }


	  ofs << "</Patch> \n";

	}


      ofs << "</Channel>" << endl;
    }



  ofs << "</ISM>\n";

  ofs.close();
  return true;

}

bool svlISM::read(const char *filename)
{



  XMLNode root = XMLNode::parseFile(filename, "ISM");
  if (root.isEmpty()) {
    SVL_LOG(SVL_LOG_WARNING, "could not read Implicit Shape Model " << filename);
    return false;
  }



  _width = atoi(root.getAttribute("width"));
  _height = atoi(root.getAttribute("height"));
  _halfWidth = atoi(root.getAttribute("halfWidth"));
  _halfHeight = atoi(root.getAttribute("halfHeight"));
  _numChannels = atoi(root.getAttribute("numChannels"));
  _phase = (svlISMPhaseType) atoi(root.getAttribute("phase"));


  XMLNode ipdNode = root.getChildNode("IPD");

  if (_ipd)
    delete _ipd;

  _ipd = svlInterestPointDetector::readXMLNode(ipdNode);

  _ownIpd = atoi(root.getAttribute("ownIpd"));

  _objects.clear();
  unsigned numFrames = atoi(root.getAttribute("numFrames"));
  _objects.resize(numFrames);

  for (unsigned i = 0; i < numFrames; i++)
    {
      XMLNode node = root.getChildNode("Object2dFrame",i);
      vector<svlObject2dFrame> & v = _objects;

      if (node.isEmpty() ) continue;
       
       v.at(i).resize(node.nChildNode("Object"));
       for (int j = 0; j < node.nChildNode("Object"); j++) {
	 XMLNode objNode = node.getChildNode("Object", j);
	 v.at(i).at(j).name = objNode.getAttribute("name");
	 v.at(i).at(j).x = atof(objNode.getAttribute("x"));
	 v.at(i).at(j).y = atof(objNode.getAttribute("y"));
	 v.at(i).at(j).w = atof(objNode.getAttribute("w"));
	 v.at(i).at(j).h = atof(objNode.getAttribute("h"));
	 if (v.at(i).at(j).w < 0.0) { 
	   v.at(i).at(j).x += v.at(i).at(j).w;
	   v.at(i).at(j).w = -v.at(i).at(j).w;
	 }
	 if (v.at(i).at(j).h < 0.0) { 
	   v.at(i).at(j).y += v.at(i).at(j).h;
	   v.at(i).at(j).h = -v.at(i).at(j).h;
	 }
	 if (objNode.getAttribute("pr")) {
	   v.at(i).at(j).pr = atof(objNode.getAttribute("pr"));
	 } else {
	   v.at(i).at(j).pr = 1.0;
	 }
       }
    }

  destroyPatches();


  _offsets.clear();
  _negatives.clear();
  _occurrances.clear();


  _patches.resize(_numChannels);
  _offsets.resize(_numChannels);
  _negatives.resize(_numChannels);
  _occurrances.resize(_numChannels);

  for (unsigned ch = 0; ch < _numChannels; ch++)
    {


      XMLNode channelNode = root.getChildNode("Channel",ch);
      unsigned n = atoi(channelNode.getAttribute("numEntries"));

      _patches.at(ch).resize(n, NULL);

      if (_phase == COUNT)
	{
	  _offsets.at(ch).resize(n);
	  _negatives.at(ch).resize(n);
	  _occurrances.at(ch).resize(n);
	}


      for (unsigned i = 0; i < n; i++)
	{
	  
	  
	  
	  
	  XMLNode patchNode = channelNode.getChildNode("Patch",i);
	  
	  
	  XMLNode imNode = patchNode.getChildNode("Pixels");
	  
	  CvMat * mat = cvCreateMat(_height,_width,CV_32FC1);
	  
	  assert(mat);
	  
	  vector<float> v;
	  
	  parseString(imNode.getText(), v);
	  
	  
	  
	  assert(v.size() == (unsigned) _width * _height);
	  
		
	  for (unsigned j = 0; j < _height; j++)
	    {
	      for (unsigned k = 0; k < _width; k++)
		{
		  cvmSet(mat,j,k,v.at(j*_width+k));
		}
	    }
	  
	  
	  
	  _patches.at(ch).at(i) = new svlISMpatch(mat);
	  
	  assert(_patches.at(ch).at(i));
	  
	  
	  
	  if (_phase == COUNT)
	    {
	      
	      
	      _negatives.at(ch).at(i) = atoi(patchNode.getAttribute("negatives"));
	      
	      
	      XMLNode offsetsNode = patchNode.getChildNode("Offsets");

	      const char * text = offsetsNode.getText();
	      
	      if (text)
		{
		  
		  
		  vector<int> v;
		  
		  
		  parseString(text, v);
		  
		  
		  
		  if (v.size() %2 != 0)
		    {
		      
		      
		      for (unsigned j = 0; j < v.size(); j++)
			cout << v.at(j) << endl;
		    }
		  
		  
		  assert(v.size() %2 == 0);
		  
		  
		  _offsets.at(ch).at(i).resize(v.size()/2);
		  
		  
		  for (unsigned j = 0; j < v.size(); j+=2)
		    {
		      
		      _offsets.at(ch).at(i).at(j/2) = cvPoint(v.at(j),v.at(j+1));
		    }
		  
		}
		  
	      XMLNode occurrancesNode = patchNode.getChildNode("Occurrances");
	      
	      text = occurrancesNode.getText();
	      
	      if (text)
		{
		  
		  
		  vector<int> v;
		  
		      
		  parseString(text, v);
		  
		  
		  
		  if (v.size() %3 != 0)
		    {
		      
		      for (unsigned j = 0; j < v.size(); j++)
			cout << v.at(j) << endl;
		    }
		  
		  
		  assert(v.size() %3 == 0);
		  
		  
		  _occurrances.at(ch).at(i).resize(v.size()/3);
		  
		  
		  for (unsigned j = 0; j < v.size(); j+=3)
		    {
		      
		      _occurrances.at(ch).at(i).at(j/3).x = v.at(j);
		      _occurrances.at(ch).at(i).at(j/3).y = v.at(j+1);
		      _occurrances.at(ch).at(i).at(j/3).frame = v.at(j+2);
		    }
		}
	    }  
	}
    }
  

  _ngcThreshold = atof(root.getAttribute("ngcThreshold"));

  _votes.clear();
  _voteWeights.clear();
  _voteSources.clear();

  _showInterestPointsAccum = atoi(root.getAttribute("showInterestPointsAccum"));



  return true;
}


void svlISM::destroyPatches()
{
  

  for (unsigned ch = 0; ch < _patches.size(); ch++)
    {
      for (unsigned i = 0; i < _patches.at(ch).size(); i++)
	{	  
	  delete _patches.at(ch).at(i);
	}
    }


  _patches.clear();

}

bool svlISM::operator==(const svlISM & o) const
{

  if (_objects.size() != o._objects.size())
    return false;

  for (unsigned i = 0; i < _objects.at(i).size(); i++)
    {
      if (_objects.at(i).size() != o._objects.at(i).size())
	return false;

      for (unsigned j = 0; j < _objects.at(i).size(); j++)
	{
	  if (_objects.at(i).at(j) != o._objects.at(i).at(j))
	    return false;
	}
    }


  if (_width != o._width)
    return false;

  if (_height != o._height)
    return false;

  if (_halfWidth != o._halfWidth)
    return false;

  if (_halfHeight != o._halfHeight)
    return false;

  if (_numChannels != o._numChannels)
    return false;

  if (_phase != o._phase)
    return false;

  if ( (*_ipd) != (*o._ipd))
    return false;

  if (_patches.size() != o._patches.size())
    return false;

  for (unsigned i = 0; i < _patches.size(); i++)
    {
      if (_patches.at(i).size() != o._patches.at(i).size())
	return false;

      for (unsigned j = 0; j < _patches.at(i).size(); j++)
	{
	  const svlISMpatch & p = *(_patches.at(i).at(j));
	  const svlISMpatch & op = *(o._patches.at(i).at(j));
	  if (p != op)
	    return false;
	}
    }



  if (_offsets.size() != o._offsets.size())
    return false;

  if (_occurrances.size() != o._occurrances.size())
    return false;

  for (unsigned i = 0; i < _offsets.size(); i++)
    {

      if (_offsets.at(i).size() != o._offsets.at(i).size())
	return false;


      for (unsigned j = 0; j < _offsets.at(i).size(); j++)
	{
	  if (_offsets.at(i).at(j).size() != o._offsets.at(i).at(j).size())
	    {
	      cout << "num offsets!" << endl;
	      cout << "_offsets.at("<<i<<").at("<<j<<").size(): " << _offsets.at(i).at(j).size() << endl;
	      cout << "o._offsets.at("<<i<<").at("<<j<<").size(): " << o._offsets.at(i).at(j).size() << endl;
	      return false;
	    }

	  for (unsigned k = 0; k < _offsets.at(i).at(j).size(); k++)
	    {
	      if (_offsets.at(i).at(j).at(k).x != o._offsets.at(i).at(j).at(k).x)
		{
		  cout << "x!" << endl;
		return false;
		}

	      if (_offsets.at(i).at(j).at(k).y != o._offsets.at(i).at(j).at(k).y)
		{
		  cout << "y!" << endl;
		return false;
		}
	    }
	}
    }


 for (unsigned i = 0; i < _occurrances.size(); i++)
    {

      if (_occurrances.at(i).size() != o._occurrances.at(i).size())
	return false;


      for (unsigned j = 0; j < _occurrances.at(i).size(); j++)
	{
	  if (_occurrances.at(i).at(j).size() != o._occurrances.at(i).at(j).size())
	    {
	      cout << "num occurrances!" << endl;
	      cout << "_occurrances.at("<<i<<").at("<<j<<").size(): " << _occurrances.at(i).at(j).size() << endl;
	      cout << "o._occurrances.at("<<i<<").at("<<j<<").size(): " << o._occurrances.at(i).at(j).size() << endl;
	      return false;
	    }

	  for (unsigned k = 0; k < _occurrances.at(i).at(j).size(); k++)
	    {
	      if (_occurrances.at(i).at(j).at(k).x != o._occurrances.at(i).at(j).at(k).x)
		{
		  cout << "x!" << endl;
		return false;
		}

	      if (_occurrances.at(i).at(j).at(k).y != o._occurrances.at(i).at(j).at(k).y)
		{
		  cout << "y!" << endl;
		return false;
		}
	    }
	}
    }

  assert(_negatives.size() == _offsets.size());

  if (_negatives.size())
    assert(_negatives.size() == _patches.size());

  if (_negatives.size() != o._negatives.size())
    return false;


  for (unsigned i = 0; i < _negatives.size(); i++)
    {
      if (_negatives.at(i).size() != o._negatives.at(i).size())
	return false;

      for (unsigned j = 0; j < _negatives.at(i).size(); j++)
	{
	  if (_negatives.at(i).at(j) != o._negatives.at(i).at(j))
	    return false;
	}
    }

  if (_ngcThreshold != o._ngcThreshold)
    return false;

  if (_showInterestPointsAccum != o._showInterestPointsAccum)
    return false;

  return true;
}

svlISM::svlISM(const char * filename)
{
  _ipd = NULL;

  read(filename);
}

void svlISM::debugPrune()
{
  _patches.resize(1);
  if (_patches.at(0).size())
    _patches.at(0).resize(1);

  if (_offsets.size())
    {
      _offsets.resize(1);
      if (_offsets.at(0).size())
	_offsets.at(0).resize(1);

      _negatives.resize(1);
      if (_negatives.at(0).size())
	_negatives.at(0).resize(1);

      _occurrances.resize(1);
      if (_occurrances.at(0).size())
	_occurrances.at(0).resize(1);
    }
}

svlObject2dFrame svlISM::testScaleVote(const vector<IplImage *> & channels)
{
  _votes.clear();
  _voteScales.clear();
  _voteWeights.clear();
  _voteSources.clear();

  for (unsigned ch = 0; ch < _numChannels; ch++)
    {

      cout << "A" << endl;

      vector<CvPoint> locations;
      vector<svlISMpatch *> impatches;

      cout << "gpal enter" << endl;

      assert(ch < channels.size());

      getPatchesAndLocations(channels.at(ch), impatches, locations);
      cout << "gpal exit" << endl;

      for (unsigned i = 0; i < locations.size(); i++)
	{
	  cout << "A1" << endl;
	  vector<unsigned> responders;
	  for (unsigned j = 0; j < _patches.at(ch).size(); j++)
	    {
	      if (svlNGC(impatches.at(i), _patches.at(ch).at(j)) > _ngcThreshold)
		{
		  responders.push_back(j);
		}
	    }
	  
	  double patchProb = 1.0 / double(responders.size());
	  
	  for (unsigned j = 0; j < responders.size(); j++)
	    {
	      unsigned p = responders.at(j);
	      
	      double numNeg = _negatives.at(ch).at(p);
	      double numPos = _offsets.at(ch).at(p).size();
	      
	      // P(location | object, patch) P(object | patch) P(patch | pixels)
	      double weight =  patchProb * (numPos / (numNeg+numPos))  / numPos;	      
	      
	      for (unsigned occ = 0; occ < _occurrances.at(ch).at(p).size(); occ++)
		{
		  int occ_x = _occurrances.at(ch).at(p).at(occ).x;
		  int occ_y = _occurrances.at(ch).at(p).at(occ).y;
		  unsigned f = _occurrances.at(ch).at(p).at(occ).frame;
		  
		  for (unsigned obj = 0; obj < _objects[f].size(); obj++)
		    {
		      _votes.push_back(cvPoint(locations[i].x-occ_x+_objects[f][obj].x+_objects[f][obj].w/2.0, locations[i].y-occ_y+_objects[f][obj].y+_objects[f][obj].h/2.0));
		      _voteScales.push_back(cvPoint(_objects[f][obj].w, _objects[f][obj].h));
		      _voteWeights.push_back(weight);
		      _voteSources.push_back(locations[i]);
		    }
		}
	    }   
	  cout << "A2" << endl;
	}
	  
      for (unsigned p = 0; p < impatches.size(); p++)
	delete impatches[p];

      cout << "B" << endl;
    }

  CvScalar yellow = cvScalar(0,255,255);
  
  IplImage * color = colorImage(channels.at(0));
  
  CvMat * voteImg = cvCreateMat(channels.at(0)->height, channels.at(0)->width, CV_32FC1);
  cvZero(voteImg);
  
  for (unsigned j = 0; j < _votes.size(); j++)
    {
      if ( 
	      (_votes.at(j).x >= 0) && 
	      (_votes.at(j).x <= color->width -1 )  && 
	       (_votes.at(j).y >= 0)  && 
	      (_votes.at(j).y < color->height))
	    {
	      //cout << _voteWeights.at(j) << endl;
	      cvmSet(voteImg,_votes.at(j).y,_votes.at(j).x, _voteWeights.at(j)+cvmGet(voteImg,_votes.at(j).y, _votes.at(j).x));
	    }
	}
      
      cout << "Minimum vote value: " << svlMin(voteImg) << endl;
      cout << "Maximum vote value: " << svlMax(voteImg) << endl;
      

      
      #define MAXFINDER_HALFWIDTH 10
      #define MAXFINDER_HALFHEIGHT 10

      
      double maxScore = 0;
      int maxI = 0;
      int maxJ = 0;

      for (int i = MAXFINDER_HALFHEIGHT; i < voteImg->rows-MAXFINDER_HALFHEIGHT; i++)
	{
	  for (int j = MAXFINDER_HALFWIDTH; j < voteImg->cols-MAXFINDER_HALFWIDTH; j++)
	    {
	      double vote = 0;

	      for (int ii = -MAXFINDER_HALFHEIGHT; ii <= MAXFINDER_HALFHEIGHT; ii++)
		{
		  for (int jj = -MAXFINDER_HALFWIDTH; jj <= MAXFINDER_HALFWIDTH; jj++)
		    {
		      vote += cvmGet(voteImg, i+ii,j+jj);
		    }
		}

	      if (vote >= maxScore)
		{
		  maxScore = vote;
		  maxI = i;
		  maxJ = j;
		}
	    }
	}

      cout << "Winning vote: " << maxScore << endl;
      


      vector<CvPoint> maxima;
      
      
      //svlFindLocalMaxima(voteImg, maxima, MAXFINDER_HALFWIDTH*2+1,MAXFINDER_HALFHEIGHT*2+1);
      maxima.push_back(cvPoint(maxJ, maxI));

      vector<double> scores;

      for (int m = 0; m < (int) maxima.size(); m++)
	{
	  double score = 0;
	  for (int i = - MAXFINDER_HALFHEIGHT; i <= MAXFINDER_HALFHEIGHT; i++)
	    for (int j = -MAXFINDER_HALFWIDTH; j <= MAXFINDER_HALFWIDTH; j++)
	      {
		if (maxima.at(m).y + i >= 0 && maxima.at(m).x+j >= 0 && maxima.at(m).y+i < voteImg->rows && maxima.at(m).x+j < voteImg->cols)
		score += cvmGet(voteImg, maxima.at(m).y+i,maxima.at(m).x+j);
	      }

	  if (score < 0.01)
	    {
	      maxima.erase(maxima.begin()+m);
	      m--;
	    }
	  else
	    scores.push_back(score);
	}
      


      double low = svlMin(voteImg);
      double high = svlMax(voteImg);

      for (int x = 0; x < color->width; x++)
	{
	  for (int y = 0; y < color->height; y++)
	    {
	      CvScalar pix = cvGet2D(color,y,x);
	      double alpha = (cvmGet(voteImg,y,x) - low) * (high-low);
	      for (unsigned j = 0; j < 3; j++)
		pix.val[j] =  alpha * yellow.val[j] + (1.0-alpha) * pix.val[j];
	      cvSet2D(color,y,x,pix);
	    }
	}

      svlObject2dFrame rval;


      for (unsigned m = 0; m < maxima.size(); m++)
	{

	  cout << "Vote for maximum "<<m<<"/"<<maxima.size()<<": "<<scores.at(m)<<endl;

	  
	  int centroid_x = maxima.at(m).x;
	  int centroid_y = maxima.at(m).y;
	  
	  IplImage * color_this = cvCloneImage(color);	  
	  
	  CvScalar red = cvScalar(0,0,255);
	  
	  
	  

	  for (int i = - 10; i <= 10; i++)
	    {
	      int x1 = centroid_x + i;
	      int y1 = centroid_y;
	    

	      if (x1 > 0 && y1 > 0 && x1 < color->width && y1 < color->height)
		cvSet2D(color_this,y1,x1,red);
	      
	      x1 = centroid_x;
	      y1 = centroid_y + i;
	      
	      if (x1 > 0 && y1 > 0 && x1 < color->width && y1 < color->height)
		cvSet2D(color_this,y1,x1,red);
	    }

	  

	  cvNamedWindow("Votes");
	  cvShowImage("Votes",color_this);

	  cvReleaseImage(&color_this);




	  IplImage * seg = colorImage(channels.at(0));
	  
	  unsigned minX = channels.at(0)->width-1;
	  unsigned maxX = 0;
	  unsigned minY = channels.at(0)->height-1;
	  unsigned maxY = 0;
	  


	  assert(_voteSources.size() == _votes.size());

	  vector<double> widths;
	  vector<double> heights;

	  for (unsigned i = 0; i < _votes.size(); i++)
	    {
	      int x = _votes.at(i).x;
	      int y = _votes.at(i).y;
	      
	      if (x >= centroid_x - MAXFINDER_HALFWIDTH && x <= centroid_x + MAXFINDER_HALFWIDTH && y >= centroid_y - MAXFINDER_HALFHEIGHT && y <= centroid_y + MAXFINDER_HALFHEIGHT)
		{
		  unsigned xsource = _voteSources.at(i).x;
		  unsigned ysource = _voteSources.at(i).y;
		  

		  widths.push_back(_voteScales[i].x);
		  heights.push_back(_voteScales[i].y);

		  cvSet2D(seg, ysource, xsource, yellow);

		  if (ysource < seg->height-1)
		    cvSet2D(seg, ysource+1, xsource, yellow);

		  if (ysource > 0)
		    cvSet2D(seg, ysource-1, xsource, yellow);

		  if (xsource < seg->width-1)
		    cvSet2D(seg, ysource, xsource+1, yellow);

		  if (xsource > 0)
		    cvSet2D(seg, ysource, xsource-1, yellow);
		  
		  if (xsource < minX)
		    minX = xsource;
		  
		  if (ysource < minY)
		    minY = ysource;
		  
		  if (xsource > maxX)
		    maxX = xsource;
		  
		  if (ysource > maxY)
		    maxY = ysource;
		}
	    }




	  cvRectangle(seg, cvPoint(minX,minY), cvPoint(maxX,maxY), yellow);


	  double minw = minElem(widths);
	  double minh = minElem(heights);

	  double maxw = maxElem(widths);
	  double maxh = maxElem(heights);

	  double meanw = mean(widths);
	  double meanh = mean(heights);

	  double x1 = maxx(0.0, centroid_x - minw / 2.0);
	  double y1 = maxx(0.0, centroid_y - minh / 2.0);
	  double x2 = minn(seg->width-1, centroid_x + minw / 2.0);
	  double y2 = minn( seg->height-1, centroid_y + minw / 2.0);

	  CvScalar blue = cvScalar(255,0,0);

	  cvRectangle(seg,cvPoint(x1,y1),cvPoint(x2,y2),blue);

	   x1 = maxx(0.0, centroid_x - maxw / 2.0);
	   y1 = maxx(0.0, centroid_y - maxh / 2.0);
	   x2 = minn(seg->width-1, centroid_x + maxw / 2.0);
	   y2 = minn( seg->height-1, centroid_y + maxw / 2.0);	 

	   cvRectangle(seg,cvPoint(x1,y1),cvPoint(x2,y2),red);
	   
	   x1 = maxx(0.0, centroid_x - meanw / 2.0);
	   y1 = maxx(0.0, centroid_y - meanh / 2.0);
	   x2 = minn(seg->width-1, centroid_x + meanw / 2.0);
	   y2 = minn( seg->height-1, centroid_y + meanw / 2.0);	


	   CvScalar green = cvScalar(0,255,0);
	   
	   cvRectangle(seg,cvPoint(x1,y1),cvPoint(x2,y2),green);	  

	  cvNamedWindow("Segmentation");
	  cvShowImage("Segmentation", seg);



	  cvWaitKey(0);
	  break;


	  //Resample for the "refined hypothesis"
	  
	  vector<CvPoint> votes2;
	  _voteWeights.clear();
	  vector<CvPoint> _voteSources2;
	  
	  for (unsigned ch = 0; ch < _numChannels; ch++)
	    {
	      vector<CvPoint> locations;
	      vector<svlISMpatch *> impatches;
	      
	      for (unsigned x = minX; x <= maxX; x++)
		{
		  for (unsigned y = minY; y <= maxY; y++)
		    {
		      locations.push_back(cvPoint(x,y));
		      impatches.push_back(getPatch(channels.at(ch),cvPoint(x,y)));
		    }
		}
	      



	      for (unsigned i = 0; i < locations.size(); i++)
		{
		  //cout << "Location: "<<i<<" of "<<locations.size()<<endl;
		  
		  vector<unsigned> responders;
		  for (unsigned j = 0; j < _patches.at(ch).size(); j++)
		    {
		      if (svlNGC(impatches.at(i), _patches.at(ch).at(j)) > _ngcThreshold)
			{
			  responders.push_back(j);
			}
		    }
		  
		  double patchProb = 1.0 / double(responders.size());
		  
		  for (unsigned j = 0; j < responders.size(); j++)
		    {
		      unsigned p = responders.at(j);
		      
		      double numNeg = _negatives.at(ch).at(p);
		      double numPos = _offsets.at(ch).at(p).size();
		      
		      // P(location | object, patch) P(object | patch) P(patch | pixels)
		      double weight =  patchProb * (numPos / (numNeg+numPos))  / numPos;

		      //cout << "I i=" <<i << " j="<<j<<" m="<<m<< endl;
		      assert( p < _offsets.at(ch).size());
		      assert( i < locations.size() );



		      bool debug = (i == 15965) && (j == 0) && (m == 37);

		      for (unsigned k = 0; k < _offsets.at(ch).at(p).size(); k++)
			{
			  if (debug)
			    cout << "I1" << endl;
			  CvPoint pt(cvPoint(_offsets.at(ch).at(p).at(k).x+locations.at(i).x, _offsets.at(ch).at(p).at(k).y+locations.at(i).y));
			  if (debug)
			    {
			      cout << "I2" << endl;
			      CvPoint * test = new CvPoint;
			      cout << "I2A" << endl;
			      delete test;
			      cout << "I2B" << endl;
			      test = new CvPoint[1000];
			      cout << "I2C" << endl;
			      delete [] test;
			      cout << "I2D" << endl;
			      cout << votes2.size() << endl;
			    }
			  votes2.push_back(pt);
			  if (debug)
			    cout << "I3" << endl;
			  _voteWeights.push_back(weight);
			  if (debug)
			    cout << "I4" << endl;
			  _voteSources2.push_back(locations.at(i));
			  if (debug)
			    cout << "I5" << endl;
			}
		      //cout << "J" << endl;
		    }
		}

	      for (unsigned p = 0; p < impatches.size(); p++)
		delete impatches[p];
	    }
	  
	  


	  cvZero(voteImg);

	  
	  for (unsigned j = 0; j < votes2.size(); j++)
	    {
	      if ( 
		  (votes2.at(j).x >= 0) && 
		  (votes2.at(j).x <= color->width -1 )  && 
		  (votes2.at(j).y >= 0)  && 
		  (votes2.at(j).y < color->height))
		{
		  //cout << _voteWeights.at(j) << endl;
		  cvmSet(voteImg,votes2.at(j).y,votes2.at(j).x, _voteWeights.at(j)+cvmGet(voteImg,votes2.at(j).y, votes2.at(j).x));
		}
	    }
	  
	  cout << "Minimum refined vote value: " << svlMin(voteImg) << endl;
	  cout << "Maximum refined vote value: " << svlMax(voteImg) << endl;
	  
	  
	  double maxScore = 0;
	  int maxI = 0;
	  int maxJ = 0;
	  
	  for (int i = MAXFINDER_HALFHEIGHT; i < voteImg->rows-MAXFINDER_HALFHEIGHT; i++)
	    {
	      for (int j = MAXFINDER_HALFWIDTH; j < voteImg->cols-MAXFINDER_HALFWIDTH; j++)
		{
		  double vote = 0;
		  
		  for (int ii = -MAXFINDER_HALFHEIGHT; ii <= MAXFINDER_HALFHEIGHT; ii++)
		    {
		      for (int jj = -MAXFINDER_HALFWIDTH; jj <= MAXFINDER_HALFWIDTH; jj++)
			{
			  vote += cvmGet(voteImg, i+ii,j+jj);
			}
		    }
		  
		  if (vote >= maxScore)
		    {
		      maxScore = vote;
		      maxI = i;
		      maxJ = j;
		    }
		}
	    }
	  

	  cout << "S" << endl;


	  cout << "Winning vote, refined : " << maxScore << endl;
	  
	  IplImage * color2 = cvCloneImage(color);
	  
	  cvZero(color2);
	  
	  scaleToRange(voteImg, 0, 0.75);
	  
	  for (int x = 0; x < color2->width; x++)
	    {
	      for (int y = 0; y < color2->height; y++)
		{
		  CvScalar pix = cvGet2D(color2,y,x);
		  double alpha = cvmGet(voteImg,y,x);
		  for (unsigned j = 0; j < 3; j++)
		    pix.val[j] =  alpha * yellow.val[j] + (1.0-alpha) * pix.val[j];
		  cvSet2D(color2,y,x,pix);
		}
	    }



	  
	  for (int i = - 10; i <= 10; i++)
	    {
	      int x1 = maxJ + i;
	      int y1 = maxI;
	      
	      if (x1 > 0 && y1 > 0 && x1 < color->width && y1 < color->height)
		cvSet2D(color2,y1,x1,red);
	      
	      x1 = maxJ;
	      y1 = maxI + i;
	      
	      if (x1 > 0 && y1 > 0 && x1 < color->width && y1 < color->height)
		cvSet2D(color2,y1,x1,red);
	    }
	  
	  cvNamedWindow("Refined votes");
	  cvShowImage("Refined votes",color2);
	  cvReleaseImage(&color2);
	  
	  

	  
	  
	  
	  IplImage * seg2 = colorImage(channels.at(0));
	  
	  
	  minX = channels.at(0)->width-1;
	  maxX = 0;
	  minY = channels.at(0)->height-1;
	  maxY = 0;
	  
	  assert(votes2.size() == _voteSources2.size());

	  for (unsigned i = 0; i < votes2.size(); i++)
	    {
	      int x = votes2.at(i).x;
	      int y = votes2.at(i).y;
	      
	      
	      
	      if (x >= maxJ - MAXFINDER_HALFWIDTH && x <= maxJ + MAXFINDER_HALFWIDTH && y >= maxI - MAXFINDER_HALFHEIGHT && y <= maxI + MAXFINDER_HALFHEIGHT)
		{
		 unsigned xsource = _voteSources2.at(i).x;
		 unsigned ysource = _voteSources2.at(i).y;
		  
		  cvSet2D(seg2, ysource, xsource, yellow);
		  
		  if (xsource < minX)
		    minX = xsource;
		  
		  if (ysource < minY)
		    minY = ysource;
		  
		  if (xsource > maxX)
		    maxX = xsource;
		  
		  if (ysource > maxY)
		    maxY = ysource;
		}
	    }
	  
	  cvNamedWindow("Refined Segmentation");
	  cvShowImage("Refined Segmentation", seg2);
	  
	  cvReleaseImage(&seg2);

#if CONDITIONAL_VOTE
	  assert(false);
	  CvMat * condProb = cvCreateMat(h/DOWNSCALE_VOTE,w/DOWNSCALE_VOTE,CV_32FC1);
	  IplImage * condProbImg = cvCreateImage(cvSize(w/DOWNSCALE_VOTE,h/DOWNSCALE_VOTE),IPL_DEPTH_8U,3);
	  
	  for (int i = 0; i < h / DOWNSCALE_VOTE; i++)
	    for (int j = 0; j < w / DOWNSCALE_VOTE; j++)
	      {
		double p = cvmGet(posVotes,i,j);
		double n = cvmGet(negVotes,i,j);
		double prob = p/(p+n);
		cvmSet(condProb,i,j, prob);
		
		CvScalar s = cvScalar(0,0,0);
		
		if (prob >= 0.5)
		  s.val[1] = (prob - 0.5) * 2.0 * 255.0;
		else
		  s.val[2] = (1.0-prob *2.0) * 255.0;
		
		cvSet2D(condProbImg,i,j,s);
	      }
	  
	  
	  resizeInPlace(&condProbImg, h, w, CV_INTER_NN);
	  
	  
	  
	  cvNamedWindow("Conditional Prob");
	  cvShowImage("Conditional Prob",condProbImg);
	  
	  
	  cvReleaseImage(&condProbImg);
	  cvReleaseMat(&posVotes);
	  cvReleaseMat(&negVotes);
	  cvReleaseMat(&condProb);
#endif
	  
	  cout << "Waiting" << endl;
	  //cvWaitKey(0);
	  cout << "Running" << endl;
	  

	  cvReleaseImage(&seg);
	  

	  svlObject2d obj(minX,minY,maxX-minX,maxY-minY,maxScore);

	  if (obj.w >= 500)
	    {
	      cout << "maxX: " << maxX << endl;
	      cout << "minX: " << minX << endl;
	    }

	  if (obj.h >= 300)
	    {
	      cout << "maxY: " << maxY << endl;
	      cout << "minY: " << minY << endl;
	    }


	  assert(obj.w < 500);
	  assert(obj.h < 300);

	  rval.push_back(obj);
	}

	  cvReleaseMat(&voteImg);
      cvReleaseImage(&color);


      return rval;
}


/*
TODO-- was working on this method
void svlISM::getModes()
{
  _modeCoords.clear();
  _modeSizes.clear();

  if (_votes.size() == 0)
    return;


  
  int minX = _votes[0].x;
  int maxX = _votes[0].x;
  int minY = _votes[0].y;
  int maxY = _votes[0].y;

  for (unsigned i = 1; i < _votes.size(); i++)
    {
      if (_votes[i].x > maxX)
	maxX = _votes[i].x;

      if (_votes[i].y > maxY)
	maxY = _votes[i].y;

      if (_votes[i].x < minX)
	minX = _votes[i].x;

      if (_votes[i].y < minY)
	minY = _votes[i].y;
    }



  vector<vector<vector<CvPoint> > > sorted(maxX-minX+1,vector<vector<CvPoint> >(maxY-minY+1, vector<CvPoint>() ));

  for (unsigned i = 0; i < _votes.size(); i++)
    {
      sorted[_votes[i].x-minX][_votes[i].y-minY].push_back(  size of vote i );
looks like I need a second sorted to store the weights
    }
}

*/
