/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlSortVector.h
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**   Indexed vector for sorting and recovering the permutation. Similar to
**   MATLAB [vals_sorted,idx] = sort(vals);
*****************************************************************************/

#pragma once

#include <functional>
#include <vector>
#include <iostream>
#include <algorithm>
#include <cassert>

using namespace std;

template<class V = double, class I = int>
class svlSortVector {
public:
	vector<pair<V,I> > _myvec; // made public to make iterators easy to access manually
private:
	I _ix;
	bool _sorted, _ascending;
	vector<I> _orig;

public:
	svlSortVector();
	svlSortVector(bool ascending);
	svlSortVector(int reserved_size);
	svlSortVector(int reserved_size, bool ascending);
	virtual ~svlSortVector();

	int size();
	void push_back(V val);
	void pop_back();
	void clear();
	void sort();
	void reorder();
	bool is_sorted();
	void reserve(int reserved_size);
	V get_value(I _ix);
	I get_ix(I _ix);
	I get_orig_ix(I _ix);
	V get_orig_val(I _ix);

	void setAscending(bool ascending);
	bool getAscending();
	void setDescending(bool descending);
	bool getDescending();

	// accessors
	V at(I _ix);
	I val_at(I _ix);
	I ix_at(I _ix);
	pair<V,I>& operator[](I _ix);
	pair<V,I>& back();

	void erase(int i);
	
	bool operator()(const pair<V,I> &a, const pair<V,I> &b) {
		return ( !_ascending ? a.first < b.first : a.first > b.first );
	}
};

template<class V, class I>
svlSortVector<V,I>::svlSortVector() : _sorted(false), _ix(0), _ascending(true) {
}

template<class V, class I>
svlSortVector<V,I>::svlSortVector(bool _ascending) : _sorted(false), _ix(0), _ascending(_ascending) {
}

template<class V, class I>
svlSortVector<V,I>::svlSortVector(int reserved_size) : _sorted(false), _ix(0), _ascending(true) {
	_myvec.reserve(reserved_size);
}

template<class V, class I>
svlSortVector<V,I>::svlSortVector(int reserved_size, bool _ascending) : _sorted(false), _ix(0), _ascending(_ascending) {
	_myvec.reserve(reserved_size);
}

template<class V, class I>
svlSortVector<V,I>::~svlSortVector() {
	_myvec.clear();
}

template<class V, class I>
bool svlSortVector<V,I>::is_sorted() {
	return _sorted;
}

template<class V, class I>
void svlSortVector<V,I>::reserve(int reserved_size) {
	_myvec.reserve(reserved_size);
}

template<class V, class I>
V svlSortVector<V,I>::get_value(I ix) {
	return _myvec[ix].first;
}

template<class V, class I>
I svlSortVector<V,I>::get_ix(I ix) {
	return _myvec[ix].second;
}

template<class V, class I>
I svlSortVector<V,I>::get_orig_ix(I ix) {
	return _orig[ix];
}

template<class V, class I>
V svlSortVector<V,I>::get_orig_val(I ix) {
	if ( _sorted ) {
		assert(_orig[ix]!=-1);
		return _myvec[_orig[ix]].first;
	} else {
		return _myvec[ix].first;
	}
}

template<class V, class I>
void svlSortVector<V,I>::setAscending(bool ascending) {
	_ascending = ascending;
}

template<class V, class I>
bool svlSortVector<V,I>::getAscending() {
	return _ascending;
}

template<class V, class I>
void svlSortVector<V,I>::setDescending(bool descending) {
	_ascending = !descending;
}

template<class V, class I>
bool svlSortVector<V,I>::getDescending() {
	return !_ascending;
}

template<class V, class I>
int svlSortVector<V,I>::size() {
	return (int)_myvec.size();
}

template<class V, class I>
void svlSortVector<V,I>::push_back(V val) {
	_myvec.push_back(make_pair(val,_ix));
	++_ix;
	_sorted = false;
}

template<class V, class I>
void svlSortVector<V,I>::pop_back() {
	_orig[_myvec.back().second] = -1; // invalidate original indexing
	_myvec.pop_back();
}

template<class V, class I>
void svlSortVector<V,I>::clear() {
	_myvec.clear();
	_ix = 0;
	_sorted = false;
}

template<class V, class I>
void svlSortVector<V,I>::sort() {
    if ( _ascending ){
	std::sort(_myvec.begin(), _myvec.end());
    } else {
	greater<pair<V,I> > comp;
	std::sort(_myvec.begin(), _myvec.end(), comp);
    }
	
    _ix = (int)_myvec.size();
    _orig = vector<I>(_ix,-1);
    for ( int i=0; i<(int)_myvec.size(); ++i ) { // recovery indices
	if ( _myvec[i].second != -1 ) {
	    _orig[_myvec[i].second] = i;
	}
    }
    _sorted = true;
}

template<class V, class I>
void svlSortVector<V,I>::reorder() { // reorder the indices (ix's)
	_ix = 0;
	for ( typename vector<pair<V,I> >::const_iterator i=_myvec.begin(); i!=_myvec.end(); ++i ) {
		(i->second) = _ix;
		++_ix;
	}
}

template<class V, class I>
V svlSortVector<V,I>::at(I _ix) {
	return _myvec[_ix].first;
}

template<class V, class I>
I svlSortVector<V,I>::ix_at(I _ix) {
	return _myvec[_ix].second;
}

template<class V, class I>
I svlSortVector<V,I>::val_at(I _ix) {
	return _myvec[_ix].first;
}

template<class V, class I>
pair<V,I>& svlSortVector<V,I>::operator[](I ix) {
	return _myvec[ix];
}

template<class V, class I>
pair<V,I>& svlSortVector<V,I>::back() {
	return _myvec.back();
}

template<class V, class I>
void svlSortVector<V,I>::erase(int i) {
	if ( i >= (int)_myvec.size() ) return;
	typename vector<pair<V,I> >::iterator it = _myvec.begin();
	for ( int j=0; i<j; ++j, ++it );
	_myvec.erase(it);
	_orig[it->second] = -1; // invalidate original indexing
	// does not invalidate the state of _sorted
}

