/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlClassifierResponseList.cpp
** AUTHOR(S):   Olga Russakovsky <olga@cs.stanford.edu>
**              based on applications by Stephen Gould <sgould@stanford.edu>
**
*****************************************************************************/

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include <sys/types.h>
#include <limits>

#include "svlBase.h"
#include "svlVision.h"

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#define _CRT_SECURE_NO_DEPRECATE
#include "win32/dirent.h"
#define M_LN2 0.69314718055994530942
#else
#include <dirent.h>
#endif

#define VERBOSE
#undef DEBUG
//#define DEBUG

using namespace std;


void insertScore(svlClassifierResponseList &responses, 
		 float score, bool correct) {
  responses.push_back(make_pair(score, correct ? 1.0f : 0.0f));
}

void insertScore(svlClassifierResponseList &responses, 
		 float score, float weight) {
  responses.push_back(make_pair(score, weight));
}

void normalizeScores(svlClassifierResponseList &responses) {
  int numPositives = 0;
  for (unsigned i = 0; i < responses.size(); i++)
    numPositives += (responses[i].second > 0 ? 1 : 0);

  int numNegatives = responses.size() - numPositives;

  float positiveWeight = (float)numNegatives / (float)numPositives;
  for (unsigned i = 0; i < responses.size(); i++) {
    responses[i].second = (responses[i].second > 0) ? positiveWeight : 0;
  }
}



//------------------------------------------------------------------------

inline int isNegDet(float score) {
  return (score == 0.0 ? 1 : 0);
}
inline int isPosDet(float score) {
  return (score > 0) ? 1 : 0;
}

// negative objects always contribute a score of 1; positive objects,
// on the other hand, mind contribute more or less than that depending
// on the number of +/- examples in the training set
// (so if there are 10 times more negatives than positives,
// positive objects will contribute 10 points each) 
void createPRcurve(const svlClassifierResponseList &responses,
		   svlPRcurve &curve) {
  curve.clear();

  vector<pair<float, float> > points = responses;

  if (points.size() == 0) return;

  // sort scores
  sort(points.begin(), points.end());

  vector<float> cumsumDetections(points.size(), points[0].second);
  vector<int> cumsumNegatives(points.size(), isNegDet(points[0].second));
  for (unsigned i = 1; i < cumsumDetections.size(); i++) {
    cumsumDetections[i] = cumsumDetections[i - 1] + points[i].second;
    cumsumNegatives[i] = cumsumNegatives[i-1] + isNegDet(points[i].second);
  }
  
#if 0
  SVL_LOG(SVL_LOG_DEBUG, it->first << " scores:");
  for (unsigned i = 0; i < it->second.size(); i++) {
    SVL_LOG(SVL_LOG_DEBUG, it->second[i].first << " "
	    << it->second[i].second << " " << cumsumDetections[i]);
  }
  
  SVL_LOG(SVL_LOG_DEBUG, "total " << it->first << ": " << toString(cumsumDetections.back()));
#endif

  // what happens at threshold of points[0].first (so if we output _all_ the classifications)
  float totalPositives = cumsumDetections.back();
  float totalNegatives = cumsumNegatives.back();
  float threshold = points[0].first;

  // the total (weighted) number of detections
  //float totalDetections = (float)cumsumDetections.size();
  float totalDetections = totalPositives + totalNegatives;

  float recall = 1;
  float pr = totalPositives / totalDetections;

  // if threshold == 0, that means that the first detection actually was a groundtruth object,
  // not a detection by the classifier, so shouldn't count. 
  if (threshold > 0)
    curve.push_back(svlPRpoint(threshold, recall, pr));

  // now look at each object in turn
  int indx = 1;
  while (indx < (int)points.size()) {
    if (points[indx].first != threshold) {
      // so suppose we only classify the objects with prob. >= points[indx].first to be
      // correct, then we'll have missed all objects with prob. < points[indx]
      float missedPositives = cumsumDetections[indx-1];
      int missedNegatives = cumsumNegatives[indx-1];
      float detections = totalDetections - missedPositives - missedNegatives;
      
      recall = (totalPositives - missedPositives) / totalPositives;
      pr = (totalPositives - missedPositives) / detections;
      threshold = points[indx].first;

      svlPRpoint pt(threshold, recall, pr);
      if (!pt.contains_nan()) curve.push_back(pt);
    }
    indx++;
  }
}

void createDiscretizedPRcurve(const svlClassifierResponseList &responses,
			      svlPRcurve &curve, 
			      float dthreshold) {
  curve.clear();

  vector<pair<float, float> > points = responses;

  if (points.size() == 0) return;

  // sort scores
  sort(points.begin(), points.end());

  vector<float> thresholds((unsigned)(1.0 / dthreshold));
  for (unsigned i = 0; i < thresholds.size(); i++)
    thresholds[i] = dthreshold * i;
  //thresholds[thresholds.size() - 1] = 1.0; // for rounding mistakes
  //and for uneven dividing

  // to ensure that groundtruth objects that were added with prob. 0
  // don't get included
  thresholds[0] += (float)1.0e-10;

  float totalPositives = 0;
  int totalNegatives = 0;
  for (unsigned i = 0; i < points.size(); i++) {
    totalPositives += points[i].second;
    totalNegatives += isNegDet(points[i].second);
  }

  unsigned index = 0;
  float missedPositives = 0;
  int missedNegatives = 0;
  for (unsigned i = 0; i < points.size() && index < thresholds.size(); i++) {
    // have finally reached an object that's the first one to be
    // classified as positive at this threshold
    if (points[i].first >= thresholds[index]) {
      float totalDetections = totalPositives - missedPositives + totalNegatives - missedNegatives;

      float recall = (totalPositives - missedPositives) / totalPositives;
      float pr = (totalPositives - missedPositives) / totalDetections;

      float threshold = thresholds[index];
      if (index == 0) threshold = 0; // to ensure pretty output
       
      svlPRpoint pt(threshold, recall, pr);
      if (!pt.contains_nan()) curve.push_back(pt);
      index++;
    }
    missedPositives += points[i].second;
    missedNegatives += isNegDet(points[i].second);
  }
}





//--------------------------------------------------------------------

string summary(const svlClassifierResponseList &responses, float threshold) {
  float totalPositives = 0;
  float totalDetections = 0;
  float trueDetections = 0;
  for (unsigned i = 0; i < responses.size(); i++) {
    if (responses[i].first > threshold) {
      // positive and negative examples might be weighted differently
      if (responses[i].second > 0)
	totalDetections += responses[i].second;
      else
	totalDetections += 1;

      trueDetections += responses[i].second;
    }
    totalPositives += responses[i].second;
  }
    
  float recall = trueDetections/totalPositives;
  float precision = trueDetections/totalDetections;
 
  stringstream ss;
  ss << (int)totalPositives << "\t"
     << (int)trueDetections << "\t" << (int)(totalDetections - trueDetections) << "\t"
     << setprecision(3) << recall << "\t"
     << setprecision(3) << precision << "\t"
     << setprecision(3) << (2.0 * precision * recall / (precision + recall));
  return ss.str();
}




//--------------------------------------------------------------------

static float compute_MI_given_counts(int counts[2][2]) {
  int total_count = counts[0][0] + counts[0][1] + counts[1][0] + counts[1][1];

 // computing MI(X; C) = sum_c,x p(c,x)log(p(c,x)/(p(c)*p(x)))
  float MI = 0;
  for (int x = 0; x < 2; x++) {   
    int count_of_x = counts[x][0] + counts[x][1];

    for (int c = 0; c < 2; c++) {
      int cur_count = counts[x][c];
      int count_of_c = counts[0][c] + counts[1][c];

      // 0 * log(0) = 0
      if (cur_count == 0) 
	continue;

      float pcx = (float)cur_count / (float)total_count;    // p(c,x)
      float pc = (float)count_of_c / (float)total_count; // p(c)
      float px = (float)count_of_x / (float)total_count; // p(x)
      
      MI += pcx * log(pcx / (pc * px)) / M_LN2;
    }
  }

  return MI;
}

float computeMI(const svlClassifierResponseList &responses, float threshold) {
  // counts[x][c]: x = 1[value >= threshold]
  //             : c = 1[is positive detection]
  int counts[2][2] = {{0, 0}, {0, 0}};

  int total_count = (int)responses.size();

  for (int i = 0; i < total_count; i++) {
    int x = (responses[i].first >= threshold) ? 1 : 0;
    int c = (responses[i].second > 0) ? 1 : 0;
    counts[x][c]++;
  }

  return compute_MI_given_counts(counts);
}

pair<float, float> computeMaxMI(const svlClassifierResponseList &responses) {
  // don't want to change the order of the responses
  vector<pair<float, float> > points = responses;

  // sort scores
  sort(points.begin(), points.end());

  vector<int> cumsumNeg(points.size(), isNegDet(points[0].second));
  for (unsigned i = 1; i < cumsumNeg.size(); i++) {
    cumsumNeg[i] = cumsumNeg[i-1] + isNegDet(points[i].second);
  }

  int totalNeg = cumsumNeg[cumsumNeg.size()-1];
  int totalPos = responses.size() - totalNeg;

  float threshold = -1;
  float best_MI = -1;
  float best_threshold = -1;

  // analyze all of the possible unique thresholds
  for (unsigned i = 0; i < points.size(); i++) {
    if (points[i].first == threshold) continue;

    // (i = total number of detections < threshold)
    threshold = points[i].first;

    // counts[x][c]: x = 1[value >= threshold]
    //             : c = 1[is positive detection]
    int counts[2][2];
    counts[0][0] = (i == 0) ? 0 : cumsumNeg[i-1];
    counts[0][1] = i - counts[0][0];
    counts[1][0] = totalNeg - counts[0][0];
    counts[1][1] = totalPos - counts[0][1];

    float MI = compute_MI_given_counts(counts);
    if (MI > best_MI) {
      best_MI = MI;
      best_threshold = threshold;
    }
  }

  return make_pair<float, float>(best_MI, best_threshold);
}

static svlClassifierResponseCache new_cache(int size = 0) {
  vector<int> **result = new vector<int>*[2];
  for (int i = 0; i < 2; i++) {
    result[i] = new vector<int>[2];

    if (size > 0) {
      for (int j = 0; j < 2; j++)
	result[i][j].reserve(size);
    }
  }
  return result;
}

void deleteClassifierResponseCache(svlClassifierResponseCache &cache) {
  for (int i = 0; i < 2; i++)
    delete[] cache[i];
  delete[] cache;
  cache = NULL;
}

svlClassifierResponseCache
thresholdResponsesForMaxMI(const svlClassifierResponseList &responses,
			   float *best_MI, float *best_threshold) {

  pair<float, float> maximum = computeMaxMI(responses);
  if (best_MI)         *best_MI = maximum.first;
  if (best_threshold)  *best_threshold = maximum.second;

  svlClassifierResponseCache result = new_cache(responses.size());
  float threshold = maximum.second;
  for (unsigned n = 0; n < responses.size(); n++) {
    int i = (responses[n].first >= threshold) ? 1 : 0;
    int j = (responses[n].second > 0) ? 1 : 0;
    result[i][j].push_back(n);
  }

  return result;
}

// counts[x][y][c]: x = 1[feature X fires]
//                  y = 1[feature Y fires]
//                : c = 1[is positive detection]
static float compute_joint_MI_given_counts(int counts[2][2][2]) {
  int total_count = 0;
  for (int x = 0; x < 2; x++) {
    for (int y = 0; y < 2; y++) {
      for (int c = 0; c < 2; c++) {
	total_count += counts[x][y][c];
      }
    }
  }

 // computing MI(X, Y; C) = -sum_c,x,y p(c,x,y)log(p(c,x,y)/(p(c)*p(x,y)))
  float MI = 0;
  for (int x = 0; x < 2; x++) {
    for (int y = 0; y < 2; y++) {
      int count_of_xy = counts[x][y][0] + counts[x][y][1];
   
      for (int c = 0; c < 2; c++) {
	int count_of_c = counts[0][0][c] + counts[0][1][c] + counts[1][0][c] + counts[1][1][c];
	int cur_count = counts[x][y][c];

	// 0 * log(0) = 0
	if (cur_count == 0) 
	  continue;

	float pcxy = (float)cur_count / (float)total_count;    // p(c,x,y)
	float pxy = (float)count_of_xy / (float)total_count;   // p(x,y)
	float pc = (float)count_of_c / (float)total_count;     // p(c)

	//if (count_of_xy == 0)
	//SVL_LOG(SVL_LOG_VERBOSE, "\tfor x = " << x << ", y = " << y << ", the count is 0");

	//if (cur_count == 0)
	//SVL_LOG(SVL_LOG_VERBOSE, "\tfor x = " << x << ", y = " << y << ", c = " << c << ", the count is 0");

	MI += pcxy * log(pcxy / (pc * pxy)) / M_LN2;
      }
    }
  }

  return MI;
}

float computeJointMI(const svlClassifierResponseList &responses_X,
		     const svlClassifierResponseList &responses_Y) {

  if(responses_X.size() != responses_Y.size())
    SVL_LOG(SVL_LOG_FATAL, "The responses_X and responses_Y don't correspond to the same examples");
  
  // counts[x][y][c]: x = 1[feature X fires]
  //                  y = 1[feature Y fires]
  //                : c = 1[is positive detection]
  int counts[2][2][2] = {{{0, 0}, {0, 0}}, {{0, 0}, {0, 0}}};

  for (unsigned i = 0; i < responses_X.size(); i++) {
    int x = (int)responses_X[i].first;
    int y = (int)responses_Y[i].first;
    int c = (int)responses_X[i].second;
    int c2 = (int)responses_Y[i].second;
    if (c != c2) 
      SVL_LOG(SVL_LOG_FATAL, "The responses_X and responses_Y don't correspond to the same examples");
    counts[x][y][c]++;
  }

  return compute_joint_MI_given_counts(counts);
}


float computeJointMI(const svlClassifierResponseCache &cache_X,
		     const svlClassifierResponseCache &cache_Y) {
  
  // counts[x][y][c]: x = 1[feature X fires]
  //                  y = 1[feature Y fires]
  //                : c = 1[is positive detection]
  int counts[2][2][2];

  vector<int> result;
  for (int x = 0; x < 2; x++) {
    for (int y = 0; y < 2; y++) {
      for (int c = 0; c < 2; c++) {
	result.clear();
	set_intersection(cache_X[x][c].begin(), cache_X[x][c].end(),
			 cache_Y[y][c].begin(), cache_Y[y][c].end(),
			 inserter(result, result.begin()));
	counts[x][y][c] = result.size();
      }
    }
  }

  return compute_joint_MI_given_counts(counts);
}











