/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlPatchUtils.h
** AUTHOR(S):   Ian Goodfellow <ia3n@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**             
** DESCRIPTION:
**   Commonly re-used subroutines for processing command line arguments related
**   to patch types, mapping from channel types to patch types, or reading
**   and writing image files in the patch based object detector pipeline
**
*****************************************************************************/


#pragma once


#include "svlVision.h"
#include "svlSoftEdgeMap.h"


enum svlChannelType
  {
    SVL_CHANNEL_INTENSITY,
    SVL_CHANNEL_EDGE,
    SVL_CHANNEL_DEPTH,
    SVL_CHANNEL_ABS,
  };

struct svlMaskedFrame
{
svlMaskedFrame() : mask(NULL) {}
  
  vector<IplImage *> channels;
  CvMat * mask;
};



//TODO-- parse more command line options,  resize size, etc. (go through the apps that have
//already been converted, so far just the training set generator, and move over anything that could be code shared)

//TODO-- have this class print out the usage for the command line options that it handles


//Flags for the svlPatchUtil constructor (can be |ed together)
#define SVL_PATCH_DEFAULT_EDGE 1   //If no channels are specified, include an edge map by default
#define SVL_PATCH_CLI_DEPTHCHANNEL 2 //Enables the command line argument "-depthChannel"
#define SVL_PATCH_CLI_BASE_EXT 4 //Enables the command line argument "-baseExt"
#define SVL_PATCH_CLI_MASK 8 //Enables the command line argument "-mask"

class svlPatchUtil
{
 public:
  svlPatchUtil(unsigned flags = 0);
  svlPatchUtil(const svlPatchUtil & other);

  //Parses a single command line flag and its arguments. If the next argument is not one handled by the class,
  //returns false and does not touch argc or args. Otherwise, advances args to point to the next flag and
  //reduces argc appropriately
  bool parseCommandLineItem(int & argc, char ** & args);
  
  //Sets the base extension. Usually this comes from calling getExtension on an image sequence you are going to process.
  //If you are going to set the base extension, you should do it before calling finishCommandLine.
  void setBaseExt(const string & ext) { _baseExt = ext; }

  //Does any backpatching and filling in of default values necessary after all command line arguments have been read. Completes the definitions of the channels
  void finishCommandLine();

  //Uses the channel definitions to load all of the appropriate filepaths for a directory. Call only after finishCommandLine
  void readDir(const char * dir);
  
  //Gets all of the multi-channel patches referred to by the filepaths already loaded internally from a call to readDir. on return, output[i][j] is frame i, channel j
  void getAllFrames(vector<vector<IplImage *> > & output, unsigned maxNumToRead = UINT_MAX);

  //Same thing but allowing for masks
  void getAllFrames(vector<svlMaskedFrame > & output) ;

  //Gets a single specific frame. Again, call only after calling readDir
  //Return type is true on success.
  //On failure, return type is false and output is empty (cleared)
  bool getFrame(unsigned idx, vector<IplImage *> & output) ;
  bool getFrame(unsigned idx, svlMaskedFrame & output) ;

  //Like getFrame, but uses a vector of filenames to use for each channel,
  //rather than using an index into filenames obtained from readDir
  bool getFrameFromFiles(const vector<string> & files, vector<IplImage *> & output);

  //Like getFrame, but takes the name of a file that has extension _baseExt, and
  //loads all channels by swapping the extension for each channel
  bool getFrameFromBasename(const string & baseName, vector<IplImage *> & output) ;

  // Before there used to be a function loadPatch which would take an
  // svlPatchDefinitionType as the second argument. However, there are
  // now a variety of different loading schemes (and color
  // conversions), so there was unfortunately no way to get around
  // changing the second argument to be the specific channel; the new
  // function name is there just to emphasize this change
  IplImage *loadFrame(const string &path, unsigned channel) const;

 public:
  //Sets the size of the depth map
  void setDepthMapSize(int width, int height);

  //Sets the size images are resized to for getFrame/getAllFrames
  //If the size of the depth map is invalid, this also sets the depth map size
  void setResizeSize(int width, int height);

  //Frames access function
  unsigned numFrames() const;

  //Channel definition access functions
  unsigned numChannels() const { return (int)_channelTypes.size(); }
  svlColorType colorType(unsigned idx) const { return _colorTypes[idx]; }
  svlChannelType channelTypes(unsigned idx) const { return _channelTypes[idx]; }
  svlPatchDefinitionType patchTypes(unsigned idx) const { return _channelPatchTypes[idx]; }
  string extensions(unsigned idx) const { return _channelExts[idx]; }

 protected:

  static string stripExtension(const string & str, const string & ext);


  protected:

  // !!!!!!!!!!!! if you add new fields, also add them to the copy constructor  !!!!!!!!!!!!!!!!!!!!!!!!
  //todo-- is there any way of automatedly checking that all fields are handled by the copy constructor?

  //Definitions of the channels
  vector<svlChannelType> _channelTypes;
  vector<svlColorType> _colorTypes;
  vector<svlPatchDefinitionType> _channelPatchTypes;
  vector<string> _channelExts;

  //Maps text strings (ie, from the command line) to channel and color types
  map<string,svlChannelType> _textToChannelType;
  map<string, svlColorType> _textToColorType;
	
  //Maps channel type enums to patch type enums (no, this is not a one-to-one mapping, so we do need different enums)
  map<svlChannelType, svlPatchDefinitionType> _channelTypeToPatchType;

  string _baseExt; //The CLI parser will patch any channels with extension "-" to have this extension. 
  bool _cliBaseExt; //If true, the CLI parser will allow _baseExt to be set using the -baseExt flag

  bool _defaultEdge; //If true, assumes that the second channel is an edge map if no channels are specified (the first channel is always assumed to be an intensity channel if no channels are specified)

  bool _cliMask;       //If true, then CLI parser looks for the -mask flag
  bool _mask;         //If true, then frames are associated with a mask matrix file, where 1s indicate that pixels should be treated as unobserved. Right now the only use of this is to block buildPatchDictionary from using certain image regions
  string _maskExt;    //The file extension of the mask files

  string _dir;//Remembers last call to readDir. Used to synthesize full file paths out of the lists of filenames
  svlExtensionTree _eTree;//Used to get files out of a directory and map the files to the correct channel
  vector<vector<string> > _files; //_files[i][j] contains the name of the file for frame j of channel i. if _mask is true then _files[numChannels()][j] contains the name of the mask.

  int _depthMapWidth;
  int _depthMapHeight;

  int _width;
  int _height;

  svlSoftEdgeMap _softEdgeCalculator;
  
  //For handling deprecated command line options
  string _depthMapExt;
  string _imageExt;
  bool _cliDepthChannel;
  bool _hasDepthChannel;
  bool _usedImageExt;
  bool _usedDepthMapExt;
  bool _usedDeprecatedCh;
  bool _usedNewCh;

};





